#!/usr/bin/env python
from __future__ import annotations

import logging

# import contextlib
import multiprocessing as mp
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime
from os import PathLike
from pathlib import Path
from typing import Mapping, NamedTuple, Sequence

from opera_utils import group_by_burst, group_by_date  # , get_dates
from tqdm.auto import tqdm

from dolphin import __version__, io, timeseries, utils
from dolphin._log import log_runtime, setup_logging
from dolphin.timeseries import ReferencePoint
from dolphin.workflows import CallFunc

from . import stitching_bursts, unwrapping, wrapped_phase
from ._utils import _create_burst_cfg, _remove_dir_if_empty
from .config import DisplacementWorkflow  # , TimeseriesOptions

logger = logging.getLogger(__name__)


class OutputPaths(NamedTuple):
    """Named tuple of `DisplacementWorkflow` outputs."""

    comp_slc_dict: dict[str, list[Path]]
    stitched_ifg_paths: list[Path]
    stitched_cor_paths: list[Path]
    stitched_temp_coh_file: Path
    stitched_ps_file: Path
    stitched_amp_dispersion_file: Path
    stitched_shp_count_file: Path
    unwrapped_paths: list[Path] | None
    conncomp_paths: list[Path] | None
    timeseries_paths: list[Path] | None
    tropospheric_corrections: list[Path] | None
    ionospheric_corrections: list[Path] | None
    reference_point: ReferencePoint | None


@log_runtime
def run(
    cfg: DisplacementWorkflow,
    debug: bool = False,
) -> OutputPaths:
    """Run the displacement workflow on a stack of SLCs.

    Parameters
    ----------
    cfg : DisplacementWorkflow
        [`DisplacementWorkflow`][dolphin.workflows.config.DisplacementWorkflow] object
        for controlling the workflow.
    debug : bool, optional
        Enable debug logging, by default False.

    """
    if cfg.log_file is None:
        cfg.log_file = cfg.work_directory / "dolphin.log"
    # Set the logging level for all `dolphin.` modules
    setup_logging(debug=debug, filename=cfg.log_file)
    # TODO: need to pass the cfg filename for the logger
    logger.debug(cfg.model_dump())

    if not cfg.worker_settings.gpu_enabled:
        utils.disable_gpu()
    utils.set_num_threads(cfg.worker_settings.threads_per_worker)

    try:
        grouped_slc_files = group_by_burst(cfg.cslc_file_list)
    except ValueError as e:
        # Make sure it's not some other ValueError
        if "Could not parse burst id" not in str(e):
            raise
        # Otherwise, we have SLC files which are not OPERA burst files
        grouped_slc_files = {"": cfg.cslc_file_list}

    if cfg.amplitude_dispersion_files:
        grouped_amp_dispersion_files = group_by_burst(cfg.amplitude_dispersion_files)
    else:
        grouped_amp_dispersion_files = defaultdict(list)
    if cfg.amplitude_mean_files:
        grouped_amp_mean_files = group_by_burst(cfg.amplitude_mean_files)
    else:
        grouped_amp_mean_files = defaultdict(list)

    grouped_iono_files: Mapping[tuple[datetime], Sequence[str | PathLike[str]]] = {}
    if len(cfg.correction_options.ionosphere_files) > 0:
        for fmt in cfg.correction_options._iono_date_fmt:
            group_iono = group_by_date(
                cfg.correction_options.ionosphere_files,
                file_date_fmt=fmt,
            )
            if len(next(iter(group_iono))) == 0:
                continue
            grouped_iono_files = {**grouped_iono_files, **group_iono}

    # ######################################
    # 1. Burst-wise Wrapped phase estimation
    # ######################################
    if len(grouped_slc_files) > 1:
        logger.info(f"Found SLC files from {len(grouped_slc_files)} bursts")
        wrapped_phase_cfgs = [
            (
                burst,  # Include the burst for logging purposes
                _create_burst_cfg(
                    cfg,
                    burst,
                    grouped_slc_files,
                    grouped_amp_mean_files,
                    grouped_amp_dispersion_files,
                ),
            )
            for burst in grouped_slc_files
        ]
        for _, burst_cfg in wrapped_phase_cfgs:
            burst_cfg.create_dir_tree()
        # Remove the mid-level directories which will be empty due to re-grouping
        _remove_dir_if_empty(cfg.phase_linking._directory)
        _remove_dir_if_empty(cfg.ps_options._directory)

    else:
        # grab the only key (either a burst, or "") and use that
        cfg.create_dir_tree()
        b = next(iter(grouped_slc_files.keys()))
        wrapped_phase_cfgs = [(b, cfg)]

    ifg_file_list: list[Path] = []
    temp_coh_file_list: list[Path] = []
    ps_file_list: list[Path] = []
    amp_dispersion_file_list: list[Path] = []
    shp_count_file_list: list[Path] = []
    # The comp_slc tracking object is a dict, since we'll need to organize
    # multiple comp slcs by burst (they'll have the same filename)
    comp_slc_dict: dict[str, list[Path]] = {}
    # Now for each burst, run the wrapped phase estimation
    # Try running several bursts in parallel...
    Executor = (
        ProcessPoolExecutor
        if cfg.worker_settings.n_parallel_bursts > 1
        else utils.DummyProcessPoolExecutor
    )
    mw = cfg.worker_settings.n_parallel_bursts
    ctx = mp.get_context("spawn")
    tqdm.set_lock(ctx.RLock())
    with Executor(
        max_workers=mw,
        mp_context=ctx,
        initializer=tqdm.set_lock,
        initargs=(tqdm.get_lock(),),
    ) as exc:
        fut_to_burst = {
            exc.submit(
                wrapped_phase.run,
                burst_cfg,
                debug=debug,
                tqdm_kwargs={
                    "position": i,
                },
            ): burst
            for i, (burst, burst_cfg) in enumerate(wrapped_phase_cfgs)
        }
        for fut in fut_to_burst:
            burst = fut_to_burst[fut]

            cur_ifg_list, comp_slcs, temp_coh, ps_file, amp_disp_file, shp_count = (
                fut.result()
            )
            ifg_file_list.extend(cur_ifg_list)
            comp_slc_dict[burst] = comp_slcs
            temp_coh_file_list.append(temp_coh)
            ps_file_list.append(ps_file)
            amp_dispersion_file_list.append(amp_disp_file)
            shp_count_file_list.append(shp_count)

    # ###################################
    # 2. Stitch burst-wise interferograms
    # ###################################

    # TODO: figure out how best to pick the corr size
    # Is there one best size? dependent on `half_window` or resolution?
    # For now, just pick a reasonable size
    corr_window_size = (11, 11)
    (
        stitched_ifg_paths,
        stitched_cor_paths,
        stitched_temp_coh_file,
        stitched_ps_file,
        stitched_amp_dispersion_file,
        stitched_shp_count_file,
    ) = stitching_bursts.run(
        ifg_file_list=ifg_file_list,
        temp_coh_file_list=temp_coh_file_list,
        ps_file_list=ps_file_list,
        amp_dispersion_list=amp_dispersion_file_list,
        shp_count_file_list=shp_count_file_list,
        stitched_ifg_dir=cfg.interferogram_network._directory,
        output_options=cfg.output_options,
        file_date_fmt=cfg.input_options.cslc_date_fmt,
        corr_window_size=corr_window_size,
    )

    # ###################################
    # 3. Unwrap stitched interferograms
    # ###################################
    if not cfg.unwrap_options.run_unwrap:
        logger.info("Skipping unwrap step")
        _print_summary(cfg)
        return OutputPaths(
            comp_slc_dict=comp_slc_dict,
            stitched_ifg_paths=stitched_ifg_paths,
            stitched_cor_paths=stitched_cor_paths,
            stitched_temp_coh_file=stitched_temp_coh_file,
            stitched_ps_file=stitched_ps_file,
            stitched_amp_dispersion_file=stitched_amp_dispersion_file,
            stitched_shp_count_file=stitched_shp_count_file,
            unwrapped_paths=None,
            conncomp_paths=None,
            timeseries_paths=None,
            tropospheric_corrections=None,
            ionospheric_corrections=None,
            reference_point=None,
        )

    row_looks, col_looks = cfg.phase_linking.half_window.to_looks()
    nlooks = row_looks * col_looks
    unwrapped_paths, conncomp_paths = unwrapping.run(
        ifg_file_list=stitched_ifg_paths,
        cor_file_list=stitched_cor_paths,
        temporal_coherence_file=stitched_temp_coh_file,
        nlooks=nlooks,
        unwrap_options=cfg.unwrap_options,
        mask_file=cfg.mask_file,
    )

    # ##############################################
    # 4. If a network was unwrapped, invert it
    # ##############################################

    ts_opts = cfg.timeseries_options
    # Skip if we didn't ask for inversion/velocity
    # TODO: the troposphere/ionosphere corrections rely on `timeseries_paths`
    # Perhaps we should refactor those to not need files,
    # or perhaps we should throw an error if they want corrections but not `timeseries`
    if ts_opts.run_inversion or ts_opts.run_velocity:
        # the output of run_timeseries is not currently used so pre-commit removes it
        # let's add back if we need it
        timeseries_paths, reference_point = timeseries.run(
            unwrapped_paths=unwrapped_paths,
            conncomp_paths=conncomp_paths,
            corr_paths=stitched_cor_paths,
            condition_file=stitched_temp_coh_file,
            condition=CallFunc.MAX,
            output_dir=ts_opts._directory,
            run_velocity=ts_opts.run_velocity,
            velocity_file=ts_opts._velocity_file,
            correlation_threshold=ts_opts.correlation_threshold,
            # TODO: do i care to configure block shape, or num threads from somewhere?
            # num_threads=cfg.worker_settings....?
        )

    else:
        timeseries_paths = None
        reference_point = None

    # ##############################################
    # 5. Estimate corrections for each interferogram
    # ##############################################
    tropo_paths: list[Path] | None = None
    iono_paths: list[Path] | None = None
    if len(cfg.correction_options.geometry_files) > 0:
        out_dir = cfg.work_directory / cfg.correction_options._atm_directory
        out_dir.mkdir(exist_ok=True)
        grouped_slc_files = group_by_date(cfg.cslc_file_list)

        # Prepare frame geometry files
        geometry_dir = out_dir / "geometry"
        geometry_dir.mkdir(exist_ok=True)
        assert timeseries_paths is not None
        crs = io.get_raster_crs(timeseries_paths[0])
        epsg = crs.to_epsg()
        out_bounds = io.get_raster_bounds(timeseries_paths[0])
        frame_geometry_files = utils.prepare_geometry(
            geometry_dir=geometry_dir,
            geo_files=cfg.correction_options.geometry_files,
            matching_file=timeseries_paths[0],
            dem_file=cfg.correction_options.dem_file,
            epsg=epsg,
            out_bounds=out_bounds,
            strides=cfg.output_options.strides,
        )

        # Troposphere
        if "height" not in frame_geometry_files:
            logger.warning(
                "DEM file is not given, skip estimating tropospheric corrections."
            )
        else:
            if cfg.correction_options.troposphere_files is not None:
                from dolphin.atmosphere import estimate_tropospheric_delay

                assert timeseries_paths is not None
                logger.info(
                    "Calculating tropospheric corrections for %s files.",
                    len(timeseries_paths),
                )
                tropo_paths = estimate_tropospheric_delay(
                    ifg_file_list=timeseries_paths,
                    troposphere_files=cfg.correction_options.troposphere_files,
                    file_date_fmt=cfg.correction_options.tropo_date_fmt,
                    slc_files=grouped_slc_files,
                    geom_files=frame_geometry_files,
                    reference_point=reference_point,
                    output_dir=out_dir,
                    tropo_model=cfg.correction_options.tropo_model,
                    tropo_delay_type=cfg.correction_options.tropo_delay_type,
                    epsg=epsg,
                    bounds=out_bounds,
                )
            else:
                logger.info("No weather model, skip tropospheric correction.")

        # Ionosphere
        if grouped_iono_files:
            from dolphin.atmosphere import estimate_ionospheric_delay

            logger.info(
                "Calculating ionospheric corrections for %s files",
                len(timeseries_paths),
            )
            assert timeseries_paths is not None
            iono_paths = estimate_ionospheric_delay(
                ifg_file_list=timeseries_paths,
                slc_files=grouped_slc_files,
                tec_files=grouped_iono_files,
                geom_files=frame_geometry_files,
                reference_point=reference_point,
                output_dir=out_dir,
                epsg=epsg,
                bounds=out_bounds,
            )
        else:
            logger.info("No TEC files, skip ionospheric correction.")

    # Print the maximum memory usage for each worker
    _print_summary(cfg)
    return OutputPaths(
        comp_slc_dict=comp_slc_dict,
        stitched_ifg_paths=stitched_ifg_paths,
        stitched_cor_paths=stitched_cor_paths,
        stitched_temp_coh_file=stitched_temp_coh_file,
        stitched_ps_file=stitched_ps_file,
        stitched_amp_dispersion_file=stitched_amp_dispersion_file,
        stitched_shp_count_file=stitched_shp_count_file,
        unwrapped_paths=unwrapped_paths,
        # TODO: Let's keep the unwrapped_paths since all the outputs are
        # corresponding to those and if we have a network unwrapping, the
        # inversion would create different single-reference network and we need
        # to update other products like conncomp
        # unwrapped_paths=inverted_phase_paths,
        conncomp_paths=conncomp_paths,
        timeseries_paths=timeseries_paths,
        tropospheric_corrections=tropo_paths,
        ionospheric_corrections=iono_paths,
        reference_point=reference_point,
    )


def _print_summary(cfg):
    """Print the maximum memory usage and version info."""
    max_mem = utils.get_max_memory_usage(units="GB")
    logger.info(f"Maximum memory usage: {max_mem:.2f} GB")
    logger.info(f"Config file dolphin version: {cfg._dolphin_version}")
    logger.info(f"Current running dolphin version: {__version__}")



from datetime import datetime
from pathlib import Path

import numpy as np
import numpy.testing as npt
import pytest

from dolphin import io, utils
from dolphin.interferogram import (
    Network,
    VRTInterferogram,
    _create_vrt_conj,
    estimate_correlation_from_phase,
)


def test_derived_vrt_interferogram(slc_file_list):
    """Basic test that the VRT loads the same as S1 * S2.conj()."""
    ifg = VRTInterferogram(ref_slc=slc_file_list[0], sec_slc=slc_file_list[1])

    assert ifg.path.name == "20220101_20220102.int.vrt"
    assert io.get_raster_xysize(ifg.path) == io.get_raster_xysize(slc_file_list[0])
    assert ifg.dates == (datetime(2022, 1, 1), datetime(2022, 1, 2))

    arr0 = io.load_gdal(slc_file_list[0])
    arr1 = io.load_gdal(slc_file_list[1])
    ifg_arr = ifg.load()
    assert ifg_arr.shape == arr0.shape
    npt.assert_allclose(ifg_arr, arr0 * arr1.conj(), rtol=1e-6)


def test_specify_dates(slc_file_list):
    ref_slc, sec_slc = slc_file_list[0:2]
    ref_date, sec_date = datetime(2022, 1, 1), datetime(2022, 1, 2)
    ifg = VRTInterferogram(ref_slc=ref_slc, sec_slc=sec_slc)
    assert ifg.dates == (ref_date, sec_date)
    assert ifg.path.name == "20220101_20220102.int.vrt"

    # Check other dates don't fail or get overwritten
    ref_date2 = datetime(2023, 2, 2)
    sec_date2 = datetime(2023, 2, 3)
    ifg = VRTInterferogram(
        ref_slc=ref_slc, sec_slc=sec_slc, ref_date=ref_date2, sec_date=sec_date2
    )
    assert ifg.dates == (ref_date2, sec_date2)
    assert ifg.path.name == "20230202_20230203.int.vrt"
    # One at a time check
    ifg = VRTInterferogram(ref_slc=ref_slc, sec_slc=sec_slc, ref_date=ref_date2)
    assert ifg.dates == (ref_date2, sec_date)
    assert ifg.path.name == "20230202_20220102.int.vrt"
    ifg = VRTInterferogram(ref_slc=ref_slc, sec_slc=sec_slc, sec_date=sec_date2)
    assert ifg.dates == (ref_date, sec_date2)
    assert ifg.path.name == "20220101_20230203.int.vrt"


def test_derived_vrt_interferogram_nc(slc_file_list_nc):
    ifg = VRTInterferogram(
        ref_slc=slc_file_list_nc[0], sec_slc=slc_file_list_nc[1], subdataset="data"
    )

    assert ifg.path.name == "20220101_20220102.int.vrt"
    assert io.get_raster_xysize(ifg.path) == io.get_raster_xysize(slc_file_list_nc[0])

    arr0 = io.load_gdal(slc_file_list_nc[0])
    arr1 = io.load_gdal(slc_file_list_nc[1])

    ifg_arr = ifg.load()
    assert ifg_arr.shape == arr0.shape
    npt.assert_allclose(ifg_arr, arr0 * arr1.conj(), rtol=1e-6)


def test_derived_vrt_interferogram_with_subdataset(slc_file_list_nc_with_sds):
    ifg = VRTInterferogram(
        ref_slc=slc_file_list_nc_with_sds[0], sec_slc=slc_file_list_nc_with_sds[1]
    )

    assert ifg.path.name == "20220101_20220102.int.vrt"
    assert io.get_raster_xysize(ifg.path) == io.get_raster_xysize(
        slc_file_list_nc_with_sds[0]
    )

    arr0 = io.load_gdal(slc_file_list_nc_with_sds[0])
    arr1 = io.load_gdal(slc_file_list_nc_with_sds[1])

    ifg_arr = ifg.load()
    assert ifg_arr.shape == arr0.shape
    npt.assert_allclose(ifg_arr, arr0 * arr1.conj(), rtol=1e-6)

    # Now try with just the path
    path1 = utils._get_path_from_gdal_str(slc_file_list_nc_with_sds[0])
    path2 = utils._get_path_from_gdal_str(slc_file_list_nc_with_sds[1])
    # it should fail if we don't pass the subdataset
    with pytest.raises(ValueError):
        ifg2 = VRTInterferogram(ref_slc=path1, sec_slc=path2)

    ifg2 = VRTInterferogram(
        ref_slc=path1,
        sec_slc=path2,
        path=ifg.path.parent / "test2.vrt",
        subdataset="data/VV",
    )
    ifg_arr2 = ifg2.load()
    npt.assert_allclose(ifg_arr2, ifg_arr, rtol=1e-6)


def test_derived_vrt_interferogram_outdir(tmp_path, slc_file_list):
    ifg = VRTInterferogram(ref_slc=slc_file_list[0], sec_slc=slc_file_list[1])
    assert slc_file_list[0].parent / "20220101_20220102.int.vrt" == ifg.path

    ifg = VRTInterferogram(
        ref_slc=slc_file_list[0], sec_slc=slc_file_list[1], outdir=tmp_path
    )
    assert tmp_path / "20220101_20220102.int.vrt" == ifg.path


def test_derived_vrt_interferogram_outfile(tmpdir, slc_file_list):
    # Change directory so we dont create a file in the source directory
    with tmpdir.as_cwd():
        ifg = VRTInterferogram(
            ref_slc=slc_file_list[0], sec_slc=slc_file_list[1], path="test_ifg.vrt"
        )
    assert Path("test_ifg.vrt") == ifg.path


# Use use four files for the tests below
@pytest.fixture()
def four_slc_files(slc_file_list):
    # starts on 20220101
    return slc_file_list[:4]


def _get_pair_stems(slc_file_pairs):
    return [
        (a.stem.strip(".slc.tif"), b.stem.strip(".slc.tif"))  # noqa: B005
        for a, b in slc_file_pairs
    ]


class TestNetwork:
    def test_single_reference_network(self, tmp_path, four_slc_files):
        n = Network(four_slc_files, reference_idx=0, outdir=tmp_path)

        assert n.slc_file_pairs[0][0] == four_slc_files[0]
        assert n.slc_file_pairs[0][1] == four_slc_files[1]

        assert _get_pair_stems(n.slc_file_pairs) == [
            ("20220101", "20220102"),
            ("20220101", "20220103"),
            ("20220101", "20220104"),
        ]
        # check the written out files:
        assert Path(tmp_path / "20220101_20220102.int.vrt").exists()
        assert Path(tmp_path / "20220101_20220103.int.vrt").exists()
        assert Path(tmp_path / "20220101_20220104.int.vrt").exists()

        n = Network(four_slc_files, reference_idx=1, outdir=tmp_path)
        assert _get_pair_stems(n.slc_file_pairs) == [
            ("20220101", "20220102"),  # still has the same order (early, late)
            ("20220102", "20220103"),
            ("20220102", "20220104"),
        ]
        assert Path(tmp_path / "20220101_20220102.int.vrt").exists()
        assert Path(tmp_path / "20220102_20220103.int.vrt").exists()
        assert Path(tmp_path / "20220102_20220104.int.vrt").exists()

    def test_limit_by_bandwidth(self, tmp_path, four_slc_files):
        n = Network(four_slc_files, max_bandwidth=1, outdir=tmp_path)
        assert _get_pair_stems(n.slc_file_pairs) == [
            ("20220101", "20220102"),
            ("20220102", "20220103"),
            ("20220103", "20220104"),
        ]
        n = Network(four_slc_files, max_bandwidth=2, outdir=tmp_path)
        assert _get_pair_stems(n.slc_file_pairs) == [
            ("20220101", "20220102"),
            ("20220101", "20220103"),
            ("20220102", "20220103"),
            ("20220102", "20220104"),
            ("20220103", "20220104"),
        ]
        n = Network(four_slc_files, max_bandwidth=3, outdir=tmp_path)
        assert _get_pair_stems(n.slc_file_pairs) == [
            ("20220101", "20220102"),
            ("20220101", "20220103"),
            ("20220101", "20220104"),
            ("20220102", "20220103"),
            ("20220102", "20220104"),
            ("20220103", "20220104"),
        ]
        with pytest.raises(ValueError):
            Network(four_slc_files, max_bandwidth=0)

    def test_limit_by_temporal_baseline(self, tmp_path, four_slc_files):
        with pytest.raises(ValueError):
            Network(four_slc_files, max_temporal_baseline=0)

        n = Network(four_slc_files, max_temporal_baseline=1, outdir=tmp_path)
        assert _get_pair_stems(n.slc_file_pairs) == [
            ("20220101", "20220102"),
            ("20220102", "20220103"),
            ("20220103", "20220104"),
        ]
        n = Network(four_slc_files, max_temporal_baseline=2, outdir=tmp_path)
        assert _get_pair_stems(n.slc_file_pairs) == [
            ("20220101", "20220102"),
            ("20220101", "20220103"),
            ("20220102", "20220103"),
            ("20220102", "20220104"),
            ("20220103", "20220104"),
        ]
        n = Network(four_slc_files, max_temporal_baseline=500, outdir=tmp_path)
        assert _get_pair_stems(n.slc_file_pairs) == [
            ("20220101", "20220102"),
            ("20220101", "20220103"),
            ("20220101", "20220104"),
            ("20220102", "20220103"),
            ("20220102", "20220104"),
            ("20220103", "20220104"),
        ]

    def test_annual_ifgs(self):
        dates = [
            datetime(2021, 1, 1),
            datetime(2021, 2, 1),
            datetime(2022, 1, 4),
            datetime(2022, 6, 1),
        ]
        slcs = [d.strftime("%Y%m%d") for d in dates]
        assert Network.find_annuals(slcs, dates) == [
            (slcs[0], slcs[2]),
            (slcs[1], slcs[2]),
        ]
        assert Network.find_annuals(slcs, dates, buffer_days=10) == [(slcs[0], slcs[2])]

        assert Network.find_annuals(slcs[1:], dates[1:], buffer_days=10) == []

    def test_manual_indexes(self, tmp_path, four_slc_files):
        n = Network(four_slc_files, indexes=[(0, 1)], outdir=tmp_path)
        assert _get_pair_stems(n.slc_file_pairs) == [
            ("20220101", "20220102"),
        ]

        n = Network(four_slc_files, indexes=[(0, -1), (0, 1)], outdir=tmp_path)
        assert _get_pair_stems(n.slc_file_pairs) == [
            # It should still always come back sorted
            ("20220101", "20220102"),
            ("20220101", "20220104"),
        ]

    def test_empty_error(self, four_slc_files):
        with pytest.raises(ValueError):
            Network(four_slc_files)
            Network(four_slc_files, indexes=[])
            Network(four_slc_files, max_bandwidth=0)
            Network(four_slc_files, max_temporal_baseline=0)

    def test_combination(self, four_slc_files):
        n1 = Network(four_slc_files, reference_idx=0, write=False)
        single_ref_pairs = [
            ("20220101", "20220102"),
            ("20220101", "20220103"),
            ("20220101", "20220104"),
        ]
        assert _get_pair_stems(n1.slc_file_pairs) == single_ref_pairs

        n2 = Network(
            four_slc_files, reference_idx=0, indexes=[(1, 2), (1, 3)], write=False
        )
        new_ifgs = [
            ("20220102", "20220103"),
            ("20220102", "20220104"),
        ]
        assert _get_pair_stems(n2.slc_file_pairs) == sorted(single_ref_pairs + new_ifgs)


@pytest.fixture()
def expected_3x3_cor():
    return np.ones((3, 3))


@pytest.mark.parametrize("window_size", [3, (3, 3)])
def test_correlation_from_phase_complex_input(expected_3x3_cor, window_size):
    ifg = np.arange(1, 10).reshape(3, 3).astype(np.complex64)
    result = estimate_correlation_from_phase(ifg, window_size)
    npt.assert_allclose(result, expected_3x3_cor)


@pytest.mark.parametrize("window_size", [3, (3, 3)])
def test_correlation_from_phase_square_window_phase_input(
    expected_3x3_cor, window_size
):
    phase = np.ones((3, 3))
    result = estimate_correlation_from_phase(phase, window_size)
    npt.assert_allclose(result, expected_3x3_cor)


@pytest.mark.parametrize("window_size", [3, (3, 3)])
def test_correlation_from_phase_nan_and_zero_input(expected_3x3_cor, window_size):
    ifg = np.array([[0, np.nan, 0], [1.0 + 0.0j, 1.0, np.nan], [1.0, 1.0, 1.0]])
    window_size = 3
    result = estimate_correlation_from_phase(ifg, window_size)
    expected2 = expected_3x3_cor.copy()
    expected2[0, 0] = expected2[0, 2] = 0.0
    expected2[0, 1] = expected2[1, 2] = np.nan

    npt.assert_allclose(result, expected2, equal_nan=True)


@pytest.mark.parametrize("window_size", [3, (3, 3)])
def test_correlation_from_phase_vrtinterferogram_input(window_size, slc_file_list):
    ifg = VRTInterferogram(ref_slc=slc_file_list[0], sec_slc=slc_file_list[1])
    estimate_correlation_from_phase(ifg, window_size)
    # just checking it loads and runs


def test_create_vrt_conj(tmp_path, slc_file_list_nc_wgs84):
    # create a VRTInterferogram
    infile = io.format_nc_filename(slc_file_list_nc_wgs84[0], "data")
    out = tmp_path / "test.vrt"
    _create_vrt_conj(infile, output_filename=out)

    assert out.exists()
    assert io.get_raster_gt(out) == io.get_raster_gt(infile)
    assert io.get_raster_xysize(out) == io.get_raster_xysize(infile)
    assert io.get_raster_crs(out) == io.get_raster_crs(infile)
    np.testing.assert_array_equal(io.load_gdal(out).conj(), io.load_gdal(infile))


def test_network_manual_dates(four_slc_files):
    Network(
        four_slc_files,
        max_bandwidth=1,
        write=False,
        dates=["20210101", "20210107", "20210101", "20210109"],
    )


def test_network_manual_wrong_len_dates(four_slc_files):
    with pytest.raises(ValueError):
        Network(
            four_slc_files, max_bandwidth=1, write=False, dates=["20210101", "20210109"]
        )


def test_network_no_verify():
    datestrs = ["20210101", "20210107", "20210108", "20210109"]
    Network(
        datestrs,
        max_bandwidth=1,
        write=False,
        verify_slcs=False,
    )


def test_network_from_ifgs():
    """Check that the `Network` can work when passing in ifgs"""
    ifg_files = ["20210101_20210107", "20210101_20210108", "20210101_20210109"]
    n = Network(
        ifg_files,
        max_bandwidth=10,
        write=False,
        verify_slcs=False,
        dates=["2021-01-07", "2021-01-08", "2021-01-09"],
    )
    assert len(n.ifg_list) == 3
    assert n.ifg_list[0].path.name == "20210107_20210108.int.vrt"
    assert n.ifg_list[1].path.name == "20210107_20210109.int.vrt"
    assert n.ifg_list[2].path.name == "20210108_20210109.int.vrt"



from dolphin._show_versions import _get_deps_info, _get_sys_info, show_versions


def test_get_sys_info():
    sys_info = _get_sys_info()

    assert "python" in sys_info
    assert "executable" in sys_info
    assert "machine" in sys_info


def test_get_deps_info():
    deps_info = _get_deps_info()

    assert "osgeo.gdal" in deps_info
    assert "numpy" in deps_info
    assert "pydantic" in deps_info
    assert "ruamel_yaml" in deps_info
    assert "h5py" in deps_info


def test_show_versions(capsys):
    show_versions()
    out, err = capsys.readouterr()

    assert "python" in out
    assert "dolphin" in out



from typing import Any

from .ionosphere import estimate_ionospheric_delay
from .troposphere import estimate_tropospheric_delay

__all__ = [
    "estimate_ionospheric_delay",
    "estimate_tropospheric_delay",
]


def __getattr__(name: str) -> Any:
    if name == "delay_from_netcdf":
        # let's load this module lazily to avoid an ImportError
        # for when we don't use netcdf
        from ._netcdf import delay_from_netcdf

        return delay_from_netcdf

    if name in __all__:
        return globals()[name]

    errmsg = f"module {__name__!r} has no attribute {name!r}"
    raise AttributeError(errmsg)


def __dir__() -> list[str]:
    try:
        from ._netcdf import delay_from_netcdf
    except ModuleNotFoundError:
        return __all__
    else:
        return sorted([*__all__, "delay_from_netcdf"])



import numpy as np
import numpy.testing as npt

from dolphin import utils


def test_get_types():
    np_dtype = np.dtype("complex64")
    assert utils.numpy_to_gdal_type(np_dtype) == 10
    assert np_dtype == utils.gdal_to_numpy_type(10)

    # round trip float32
    assert utils.gdal_to_numpy_type(utils.numpy_to_gdal_type(np.float32)) == np.float32


class TestTakeLooks:
    def test_basic(self):
        arr = np.array([[0.1, 0.01, 2], [3, 4, 1 + 1j]])

        downsampled = utils.take_looks(arr, 2, 1, func_type="nansum")
        npt.assert_array_equal(downsampled, np.array([[3.1, 4.01, 3.0 + 1.0j]]))
        downsampled = utils.take_looks(arr, 2, 1, func_type="mean")
        npt.assert_array_equal(downsampled, np.array([[1.55, 2.005, 1.5 + 0.5j]]))
        downsampled = utils.take_looks(arr, 1, 2, func_type="mean")
        npt.assert_array_equal(downsampled, np.array([[0.055], [3.5]]))

    def test_3d(self):
        arr = np.array([[0.1, 0.01, 2], [3, 4, 1 + 1j]])
        arr3d = np.stack([arr, arr, arr], axis=0)
        downsampled = utils.take_looks(arr3d, 2, 1)
        expected = np.array([[3.1, 4.01, 3.0 + 1.0j]])
        for i in range(3):
            npt.assert_array_equal(downsampled[i], expected)

    def test_nans(self, slc_samples):
        slc_stack = slc_samples.reshape(30, 11, 11)
        mask = np.zeros((11, 11), dtype=bool)
        # Mask the top row
        mask[0, :] = True
        slc_samples_masked = slc_stack[:, ~mask]
        s1 = np.nansum(slc_samples_masked, axis=1)

        slc_stack_masked = slc_stack.copy()
        slc_stack_masked[:, mask] = np.nan
        s2 = np.squeeze(utils.take_looks(slc_stack_masked, 11, 11))

        npt.assert_array_almost_equal(s1, s2, decimal=5)

    def test_masked_array(self):
        arr = np.ma.MaskedArray(
            [[-999, 3, 4, 1 + 1j]], mask=[[True, False, False, False]]
        )
        downsampled = utils.take_looks(arr, 1, 2, func_type="nansum")
        npt.assert_array_equal(
            downsampled, np.ma.MaskedArray([[np.inf, 5 + 1j]], mask=[True, False])
        )
        assert type(downsampled) is np.ma.MaskedArray

    def test_nomask(self):
        arr = np.ma.MaskedArray([[0.1, 0.01, 2], [3, 4, 1 + 1j]], mask=np.ma.nomask)
        downsampled = utils.take_looks(arr, 2, 1, func_type="nansum")
        npt.assert_array_equal(downsampled, np.array([[3.1, 4.01, 3.0 + 1.0j]]))
        assert type(downsampled) is np.ma.MaskedArray


def test_upsample_nearest():
    arr = np.arange(16).reshape(4, 4)
    looked = utils.take_looks(arr, 2, 2, func_type="max")
    assert looked.shape == (2, 2)
    npt.assert_array_equal(looked, np.array([[5, 7], [13, 15]]))

    upsampled = utils.upsample_nearest(looked, output_shape=arr.shape)
    assert upsampled.shape == (4, 4)
    npt.assert_array_equal(
        upsampled,
        np.array(
            [
                [5, 5, 7, 7],
                [5, 5, 7, 7],
                [13, 13, 15, 15],
                [13, 13, 15, 15],
            ]
        ),
    )

    arr3d = np.stack([arr, arr, arr], axis=0)
    looked3d = utils.take_looks(arr3d, 2, 2, func_type="max")
    assert looked3d.shape == (3, 2, 2)
    upsampled3d = utils.upsample_nearest(looked3d, output_shape=arr.shape)
    assert upsampled3d.shape == (3, 4, 4)
    for img in upsampled3d:
        npt.assert_array_equal(img, upsampled)



from __future__ import annotations

import datetime
import logging
import re
from pathlib import Path
from typing import Mapping, Sequence

import numpy as np
import opera_utils as oput
from numpy.typing import ArrayLike
from opera_utils import get_dates
from rasterio.crs import CRS
from rasterio.warp import transform_bounds
from scipy import interpolate

from dolphin import io
from dolphin._types import Bbox, Filename
from dolphin.timeseries import ReferencePoint
from dolphin.utils import format_date_pair

logger = logging.getLogger(__name__)

###########

__all__ = ["estimate_ionospheric_delay"]

# constants
K = 40.31
SPEED_OF_LIGHT = 299792458  # meters per second
EARTH_RADIUS = 6371.0088e3  # km


def estimate_ionospheric_delay(
    ifg_file_list: Sequence[Path],
    slc_files: Mapping[tuple[datetime.datetime], Sequence[Filename]],
    tec_files: Mapping[tuple[datetime.datetime], Sequence[Filename]],
    geom_files: dict[str, Path],
    reference_point: ReferencePoint | None,
    output_dir: Path,
    epsg: int,
    bounds: Bbox,
) -> list[Path]:
    """Estimate the range delay (in meters) caused by ionosphere for each interferogram.

    Parameters
    ----------
    ifg_file_list : list[Path]
        List of interferogram files.
    slc_files : Dict[datetime.date, list[Filename]]
        Dictionary of SLC files indexed by date.
    tec_files : Dict[datetime.date, list[Filename]]
        Dictionary of TEC files indexed by date.
    geom_files : list[Path]
        Dictionary of geometry files with height and incidence angle, or
        los_east and los_north.
    reference_point : tuple[int, int], optional
        Reference point (row, col) used during time series inversion.
        If not provided, no spatial referencing is performed.
    output_dir : Path
        Output directory.
    epsg : int
        the EPSG code of the input data
    bounds : Bbox
        Output bounds.

    Returns
    -------
    list[Path]
        List of newly created ionospheric phase delay corrections.

    """
    if epsg != 4326:
        left, bottom, right, top = transform_bounds(
            CRS.from_epsg(epsg), CRS.from_epsg(4326), *bounds
        )
    else:
        left, bottom, right, top = bounds

    # Frame center latitude and longitude
    latc = (top + bottom) / 2
    lonc = (left + right) / 2

    # Read the incidence angle
    if "los_east" in geom_files:
        # ISCE3 geocoded products
        los_east = io.load_gdal(geom_files["los_east"])
        los_north = io.load_gdal(geom_files["los_north"])
        inc_angle = np.arccos(np.sqrt(1 - los_east**2 - los_north**2)) * 180 / np.pi
    else:
        # ISCE2 radar coordinate
        inc_angle = io.load_gdal(geom_files["incidence_angle"])

    iono_inc_angle = incidence_angle_ground_to_iono(inc_angle)

    # frequency
    for key in slc_files:
        if "compressed" not in str(slc_files[key][0]).lower():
            one_of_slcs = slc_files[key][0]
            break

    wavelength = oput.get_radar_wavelength(one_of_slcs)
    freq = SPEED_OF_LIGHT / wavelength

    # output folder
    output_iono = output_dir / "ionosphere"
    output_iono.mkdir(exist_ok=True)

    output_paths: list[Path] = []

    for ifg in ifg_file_list:
        ref_date, sec_date = get_dates(ifg)

        name = f"{format_date_pair(ref_date, sec_date)}_ionoDelay.tif"
        iono_delay_product_name = output_iono / name

        output_paths.append(iono_delay_product_name)
        if iono_delay_product_name.exists():
            logger.info(
                "Tropospheric correction for interferogram "
                f"{format_date_pair(ref_date, sec_date)} already exists, skipping"
            )
            continue

        # The keys in slc_files do not necessarily have one date,
        # it means with the production file naming convention,
        # there will be multiple dates stored in the keys from get_dates
        # function. So in the following if we set reference_date = (ref_date, ),
        # there will be an error that it does not find the key.
        # Examples of the file naming convention for CSLC and compressed cslc is:
        # OPERA_L2_CSLC-S1_T042-088905-IW1\
        #                       _20221119T000000Z_20221120T000000Z_S1A_VV_v1.0.h5
        # OPERA_L2_COMPRESSED-CSLC-S1_T042-088905-IW1_20221107T000000Z_\
        #           20221107T000000Z_20230506T000000Z_20230507T000000Z_VV_v1.0.h5
        reference_date = next(key for key in slc_files if ref_date in key)
        # temporary fix for compressed SLCs while the required metadata i not included
        # in them. there will be modification in a future PR to add metadata to
        # compressed SLCs
        secondary_date = next(
            key
            for key in slc_files
            if sec_date in key and "compressed" not in str(slc_files[key][0]).lower()
        )

        secondary_time = oput.get_zero_doppler_time(slc_files[secondary_date][0])
        if "compressed" in str(slc_files[reference_date][0]).lower():
            # this is for when we have compressed slcs but the actual
            # reference date does not exist in the input data
            reference_time = secondary_time
        else:
            reference_time = oput.get_zero_doppler_time(slc_files[reference_date][0])

        reference_vtec = read_zenith_tec(
            time=reference_time,
            tec_file=tec_files[(ref_date,)][0],
            lat=latc,
            lon=lonc,
        )

        secondary_vtec = read_zenith_tec(
            time=secondary_time,
            tec_file=tec_files[(sec_date,)][0],
            lat=latc,
            lon=lonc,
        )

        range_delay_reference = vtec_to_range_delay(
            reference_vtec, iono_inc_angle, freq, obs_type="phase"
        )
        range_delay_secondary = vtec_to_range_delay(
            secondary_vtec, iono_inc_angle, freq, obs_type="phase"
        )

        ifg_iono_range_delay_radians = range_delay_reference - range_delay_secondary
        # Convert to meters, where positive corresponds to motion toward the satellite
        ifg_iono_range_delay = -wavelength / (4 * np.pi) * ifg_iono_range_delay_radians

        if reference_point is not None:
            ref_row, ref_col = reference_point
            ifg_iono_range_delay -= ifg_iono_range_delay[ref_row, ref_col]

        # Write 2D tropospheric correction layer to disc
        io.write_arr(
            arr=ifg_iono_range_delay,
            output_name=iono_delay_product_name,
            like_filename=ifg,
            units="meters",
        )

    return output_paths


def incidence_angle_ground_to_iono(inc_angle: ArrayLike, iono_height: float = 450e3):
    """Calibrate incidence angle on the ground surface to the ionosphere shell.

    Equation (11) in Yunjun et al. (2022, TGRS)

    Parameters
    ----------
    inc_angle: np.ndarray
        incidence angle on the ground in degrees
    iono_height: float
        effective ionosphere height in meters
        under the thin-shell assumption

    Returns
    -------
    np.ndarray
        incidence angle on the iono shell in degrees

    """
    # convert degrees to radians
    inc_angle_rad = inc_angle * np.pi / 180

    inc_angle_iono = np.arcsin(
        EARTH_RADIUS * np.sin(inc_angle_rad) / (EARTH_RADIUS + iono_height)
    )
    inc_angle_iono *= 180.0 / np.pi

    return inc_angle_iono


def read_zenith_tec(
    time: datetime.datetime, tec_file: Filename, lat: float, lon: float
) -> float:
    """Read and interpolate zenith TEC for the latitude and longitude of scene center.

    Parameters
    ----------
    time: datetime
        datetime of the acquisition
    tec_file: Filename
        path to the tec file corresponding to slc date
    lat: float
        latitude of scene center
    lon: float
        longitude of scene center

    Returns
    -------
    float
        zenith TEC of the scene center in TECU.

    """
    utc_seconds = time.hour * 3600.0 + time.minute * 60.0 + time.second

    return get_ionex_value(tec_file=tec_file, utc_sec=utc_seconds, lat=lat, lon=lon)


def vtec_to_range_delay(
    vtec: float, inc_angle: np.ndarray, freq: float, obs_type: str = "phase"
):
    """Calculate/predict the range delay in SAR from TEC in zenith direction.

    Equation (6-11) from Yunjun et al. (2022).

    Parameters
    ----------
    vtec: float
        zenith TEC in TECU
    inc_angle: np.ndarray
        incidence angle at the ionospheric shell in deg
    freq: float
        radar carrier frequency in Hz.
    obs_type: str
        given the same iono, the impact on offset (amplitude) and phase is reversed.

    Returns
    -------
    np.ndarray
        predicted range delay in meters

    """
    # ignore no-data value in inc_angle
    if isinstance(inc_angle, np.ndarray):
        inc_angle[inc_angle == 0] = np.nan

    # convert to TEC in LOS based on equation (3) in Chen and Zebker (2012)

    # Equation (26) in Bohm & Schuh (2013) Chapter 2.
    n_iono_group = 1 + K * vtec * 1e16 / freq**2

    # Equation (8) in Yunjun et al. (2022, TGRS)
    ref_angle = (
        np.arcsin(1 / n_iono_group * np.sin(inc_angle * np.pi / 180)) * 180 / np.pi
    )

    tec = vtec / np.cos(ref_angle * np.pi / 180.0)

    # calculate range delay based on equation (1) in Chen and Zebker (2012)
    range_delay = (tec * 1e16 * K / (freq**2)).astype(np.float32)

    # group delay = phase advance * -1
    if obs_type != "phase":
        range_delay *= -1.0

    return range_delay


def get_ionex_value(
    tec_file: Filename, utc_sec: float, lat: float, lon: float
) -> float:
    """Get the TEC value from input IONEX file for the input lat/lon/datetime.

    Parameters
    ----------
    tec_file: Filename
        path of local TEC file
    utc_sec: float
        UTC time of the day in seconds
    lat: float
        latitude in degrees
    lon: float
        longitude in degrees


    Returns
    -------
    float
        vertical TEC value in TECU

    """
    # time info
    utc_min = utc_sec / 60.0

    # read TEC file
    mins, lats, lons, tec_maps = read_ionex(tec_file)

    # interpolate between consecutive rotated TEC maps
    # reference: equation (3) in Schaer et al. (1998)

    ind0 = np.where((mins - utc_min) <= 0)[0][-1]
    ind1 = ind0 + 1

    lon0 = lon + (utc_min - mins[ind0]) * 360.0 / (24.0 * 60.0)
    lon1 = lon + (utc_min - mins[ind1]) * 360.0 / (24.0 * 60.0)

    tec_val0 = interpolate.griddata(
        points=(lons.flatten(), lats.flatten()),
        values=tec_maps[ind0, :, :].flatten(),
        xi=(lon0, lat),
        method="linear",
    )

    tec_val1 = interpolate.griddata(
        points=(lons.flatten(), lats.flatten()),
        values=tec_maps[ind1, :, :].flatten(),
        xi=(lon1, lat),
        method="linear",
    )

    tec_val = (mins[ind1] - utc_min) / (mins[ind1] - mins[ind0]) * tec_val0
    +(utc_min - mins[ind0]) / (mins[ind1] - mins[ind0]) * tec_val1

    return tec_val


def read_ionex(
    tec_file: Filename,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Read TEC file in IONEX format.

    Parameters
    ----------
    tec_file: Filename
        path to the TEC file in IONEX format

    Returns
    -------
    mins: np.ndarray
        1D np.ndarray in size of (num_map), time of the day in minutes
        (TEC maps are produced every few minute based on their predefined resolution,
        num_map is the the number of TEC maps produced in a day)
    lats: np.ndarray
        1D np.ndarray in size of (num_lat), latitude  in degrees
    lons: np.ndarray
        1D np.ndarray in size of (num_lon), longitude in degrees
    tec_maps: np.ndarray
        3D np.ndarray in size of (num_map, num_lat, num_lon), vertical TEC in TECU

    """

    def parse_map(tec_map, key="TEC", exponent=-1):
        tec_map = re.split(f".*END OF {key} MAP", tec_map)[0]
        tec_map = [
            np.fromstring(x, sep=" ")
            for x in re.split(".*LAT/LON1/LON2/DLON/H\\n", tec_map)[1:]
        ]
        return np.stack(tec_map) * 10**exponent

    # read IONEX file
    with open(tec_file) as f:
        fc = f.read()

        # read header
        header = fc.split("END OF HEADER")[0].split("\n")
        for line in header:
            if line.strip().endswith("# OF MAPS IN FILE"):
                num_map = int(line.split()[0])
            elif line.strip().endswith("DLAT"):
                lat0, lat1, lat_step = (float(x) for x in line.split()[:3])
            elif line.strip().endswith("DLON"):
                lon0, lon1, lon_step = (float(x) for x in line.split()[:3])
            elif line.strip().endswith("EXPONENT"):
                exponent = float(line.split()[0])

        # spatial coordinates
        num_lat = int((lat1 - lat0) / lat_step + 1)
        num_lon = int((lon1 - lon0) / lon_step + 1)
        lats = np.arange(lat0, lat0 + num_lat * lat_step, lat_step)
        lons = np.arange(lon0, lon0 + num_lon * lon_step, lon_step)

        # time stamps
        min_step = 24 * 60 / (num_map - 1)
        mins = np.arange(0, num_map * min_step, min_step)

        # read TEC and its RMS maps
        tec_maps = np.array(
            [
                parse_map(t, key="TEC", exponent=exponent)
                for t in fc.split("START OF TEC MAP")[1:]
            ],
            dtype=np.float32,
        )

    lon_2d, lat_2d = np.meshgrid(lons, lats, indexing="ij")

    return mins, lat_2d, lon_2d, tec_maps



#!/usr/bin/env python
import sys

import h5py


def create_hdf5_orbit_shell(source_file, dest_file):
    """Make a dummy OPERA CSLC file with data for baseline computation."""
    with h5py.File(source_file, "r") as src, h5py.File(dest_file, "w") as dst:
        # Copy the entire /metadata/orbit group
        orbit_group_dest = dst.require_group("/metadata/")
        data_group = dst.require_group("/data")
        wvl_group = dst.require_group(
            "/metadata/processing_information/input_burst_metadata"
        )
        src.copy("/metadata/orbit", orbit_group_dest)
        src.copy(
            "/metadata/processing_information/input_burst_metadata/wavelength",
            wvl_group,
        )

        datasets_to_copy = [
            "/data/x_coordinates",
            "/data/y_coordinates",
            "/data/projection",
        ]
        for dataset in datasets_to_copy:
            src.copy(dataset, data_group)


if __name__ == "__main__":
    try:
        src, dst = sys.argv[1:3]
    except IndexError:
        print(f"Usage: {sys.argv[0]} <src file> <dest file>")
        sys.exit(1)

    create_hdf5_orbit_shell(src, dst)



import logging
from pathlib import Path
from typing import Sequence

from opera_utils import get_dates

from dolphin import io, utils
from dolphin._types import PathOrStr
from dolphin.workflows.config import SpurtOptions

logger = logging.getLogger(__name__)

DEFAULT_OPTIONS = SpurtOptions()


def unwrap_spurt(
    ifg_filenames: Sequence[PathOrStr],
    output_path: PathOrStr,
    temporal_coherence_file: PathOrStr,
    cor_filenames: Sequence[PathOrStr] | None = None,
    mask_filename: PathOrStr | None = None,
    options: SpurtOptions = DEFAULT_OPTIONS,
    scratchdir: PathOrStr | None = None,
) -> tuple[list[Path], list[Path]]:
    """Perform 3D unwrapping using `spurt`."""
    from spurt.graph import Hop3Graph
    from spurt.io import SLCStackReader
    from spurt.workflows.emcf import (
        GeneralSettings,
        MergerSettings,
        SolverSettings,
        TilerSettings,
        compute_phasediff_deciles,
        get_bulk_offsets,
        get_tiles,
        merge_tiles,
        unwrap_tiles,
    )

    if cor_filenames is not None:
        assert len(ifg_filenames) == len(cor_filenames)
    if mask_filename is not None:
        _mask = io.load_gdal(mask_filename)

    if scratchdir is None:
        scratchdir = Path(output_path) / "scratch"
    gen_settings = GeneralSettings(
        output_folder=output_path,
        intermediate_folder=scratchdir,
        **options.general_settings.model_dump(),
    )

    tile_settings = TilerSettings(**options.tiler_settings.model_dump())
    slv_settings = SolverSettings(**options.solver_settings.model_dump())
    mrg_settings = MergerSettings(**options.merger_settings.model_dump())

    # Using default Hop3Graph
    # TODO: this is a weird hack.. if there are 15 dates, there are 14 interferograms
    # the spurt cli expects one of the filenames to be all 0s? maybe?
    # But also still expects them to be date1_date2.int.tif?
    g_time = Hop3Graph(len(ifg_filenames) + 1)
    logger.info(f"Using Hop3 Graph in time with { g_time.npoints } epochs.")

    date_str_to_file = _map_date_str_to_file(ifg_filenames)
    stack = SLCStackReader(
        slc_files=date_str_to_file,
        temp_coh_file=temporal_coherence_file,
        temp_coh_threshold=options.temporal_coherence_threshold,
    )
    # Run the workflow
    # Generate tiles
    get_tiles(stack, gen_settings, tile_settings)

    # Unwrap tiles
    unwrap_tiles(stack, g_time, gen_settings, slv_settings)

    # Compute overlap stats
    compute_phasediff_deciles(gen_settings, mrg_settings)

    # Compute bulk offsets
    get_bulk_offsets(stack, gen_settings, mrg_settings)

    # Merge tiles and write output
    unw_filenames = merge_tiles(stack, g_time, gen_settings, mrg_settings)
    # TODO: What can we do for conncomps? Anything? Run snaphu?
    conncomp_filenames: list[Path] = []

    return unw_filenames, conncomp_filenames


def _map_date_str_to_file(
    ifg_filenames: Sequence[PathOrStr], date_fmt: str = "%Y%m%d"
) -> dict[str, PathOrStr | None]:
    # Then list individual SLCs
    dates = [get_dates(f) for f in ifg_filenames]
    if len({d[0] for d in dates}) > 1:
        errmsg = "interferograms for spurt must be single reference."
        raise ValueError(errmsg)

    secondary_dates = [d[1] for d in dates]
    first_date = dates[0][0].strftime(date_fmt)
    date_strings = [utils.format_dates(d, fmt=date_fmt) for d in secondary_dates]

    date_str_to_file: dict[str, PathOrStr | None] = dict(
        zip(date_strings, ifg_filenames)
    )
    # first date - set to None
    # None is special case for reference epoch
    date_str_to_file[first_date] = None
    return date_str_to_file



import argparse
from pathlib import Path
from typing import TYPE_CHECKING, Any

from dolphin.workflows import CallFunc

if TYPE_CHECKING:
    _SubparserType = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    _SubparserType = Any


def get_parser(subparser=None, subcommand_name="timeseries") -> argparse.ArgumentParser:
    """Set up the command line interface."""
    metadata = {
        "description": "Create a configuration file for a displacement workflow.",
        "formatter_class": argparse.ArgumentDefaultsHelpFormatter,
        # https://docs.python.org/3/library/argparse.html#fromfile-prefix-chars
        "fromfile_prefix_chars": "@",
    }
    if subparser:
        # Used by the subparser to make a nested command line interface
        parser = subparser.add_parser(subcommand_name, **metadata)
    else:
        parser = argparse.ArgumentParser(**metadata)  # type: ignore[arg-type]

    # parser._action_groups.pop()
    parser.add_argument(
        "-o",
        "--output-dir",
        default=Path(),
        help="Path to output directory to store results",
    )
    parser.add_argument(
        "--unwrapped-paths",
        nargs=argparse.ZERO_OR_MORE,
        help=(
            "List the paths of all unwrapped interferograms. Can pass a "
            "newline delimited file with @ifg_filelist.txt"
        ),
    )
    parser.add_argument(
        "--conncomp-paths",
        nargs=argparse.ZERO_OR_MORE,
        help=(
            "List the paths of all connected component files. Can pass a "
            "newline delimited file with @conncomp_filelist.txt"
        ),
    )
    parser.add_argument(
        "--corr-paths",
        nargs=argparse.ZERO_OR_MORE,
        help=(
            "List the paths of all correlation files. Can pass a newline delimited"
            " file with @cor_filelist.txt"
        ),
    )
    parser.add_argument(
        "--condition-file",
        help=(
            "A file with the same size as each raster, like amplitude dispersion or"
            "temporal coherence to find reference point. default: amplitude dispersion"
        ),
    )
    parser.add_argument(
        "--condition",
        type=CallFunc,
        default=CallFunc.MIN,
        help=(
            "A condition to apply to condition file to find the reference point"
            "Options are [min, max]. default=min"
        ),
    )
    parser.add_argument(
        "--num-threads",
        type=int,
        default=5,
        help="Number of threads for the inversion",
    )
    parser.add_argument(
        "--run-velocity",
        action="store_true",
        help="Run the velocity estimation from the phase time series",
    )
    parser.add_argument(
        "--reference-point",
        type=int,
        nargs=2,
        metavar=("ROW", "COL"),
        default=(-1, -1),
        help=(
            "Reference point (row, col) used if performing a time series inversion. "
            "If not provided, a point will be selected from a consistent connected "
            "component with low amplitude dispersion or high temporal coherence."
        ),
    )
    parser.add_argument(
        "--correlation-threshold",
        type=range_limited_float_type,
        default=0.2,
        metavar="[0-1]",
        help="Pixels with correlation below this value will be masked out.",
    )

    parser.set_defaults(run_func=_run_timeseries)

    return parser


def range_limited_float_type(arg):
    """Type function for argparse - a float within some predefined bounds."""
    try:
        f = float(arg)
    except ValueError as err:
        raise argparse.ArgumentTypeError("Must be a floating point number") from err
    if f < 0 or f > 1:
        raise argparse.ArgumentTypeError(
            "Argument must be < " + str(1) + "and > " + str(0)
        )
    return f


def _run_timeseries(*args, **kwargs):
    """Run `dolphin.timeseries.run`.

    Wrapper for the dolphin.timeseries to invert and create velocity.
    """
    from dolphin import timeseries

    return timeseries.run(*args, **kwargs)


def main(args=None):
    """Get the command line arguments for timeseries inversion."""
    from dolphin import timeseries

    parser = get_parser()
    parsed_args = parser.parse_args(args)

    timeseries.run(**vars(parsed_args))


if __name__ == "__main__":
    main()



import json
import sys
import textwrap
import warnings
from io import StringIO
from itertools import repeat
from typing import Optional, TextIO, Union

from pydantic import BaseModel
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap

from dolphin._types import Filename


class YamlModel(BaseModel):
    """Pydantic model that can be exported to yaml."""

    def to_yaml(
        self,
        output_path: Union[Filename, TextIO],
        with_comments: bool = True,
        by_alias: bool = True,
        indent_per_level: int = 2,
    ):
        """Save configuration as a yaml file.

        Used to record the default-filled version of a supplied yaml.

        Parameters
        ----------
        output_path : Pathlike
            Path to the yaml file to save.
        with_comments : bool, default = False.
            Whether to add comments containing the type/descriptions to all fields.
        by_alias : bool, default = False.
            Whether to use the alias names for the fields.
            Passed to pydantic's ``to_json`` method.
            https://docs.pydantic.dev/usage/exporting_models/#modeljson
        indent_per_level : int, default = 2
            Number of spaces to indent per level.

        """
        yaml_obj = self._to_yaml_obj(by_alias=by_alias)

        if with_comments:
            _add_comments(
                yaml_obj,
                self.model_json_schema(by_alias=by_alias),
                indent_per_level=indent_per_level,
            )

        y = YAML()
        # https://yaml.readthedocs.io/en/latest/detail.html#indentation-of-block-sequences
        y.indent(
            offset=indent_per_level,
            mapping=indent_per_level,
            # It is best to always have sequence >= offset + 2 but this is not enforced
            # not following this advice might lead to invalid output.
            sequence=indent_per_level + 2,
        )
        if hasattr(output_path, "write"):
            y.dump(yaml_obj, output_path)
        else:
            with open(output_path, "w") as f:
                y.dump(yaml_obj, f)

    @classmethod
    def from_yaml(cls, yaml_path: Filename):
        """Load a configuration from a yaml file.

        Parameters
        ----------
        yaml_path : Pathlike
            Path to the yaml file to load.

        Returns
        -------
        Config
            Workflow configuration

        """
        y = YAML(typ="safe")
        with open(yaml_path) as f:
            data = y.load(f)

        return cls(**data)

    @classmethod
    def print_yaml_schema(
        cls,
        output_path: Union[Filename, TextIO] = sys.stdout,
        indent_per_level: int = 2,
    ):
        """Print/save an empty configuration with defaults filled in.

        Ignores the required `cslc_file_list` input, so a user can
        inspect all fields.

        Parameters
        ----------
        output_path : Pathlike
            Path or stream to save to the yaml file to.
            By default, prints to stdout.
        indent_per_level : int, default = 2
            Number of spaces to indent per level.

        """
        full_dict = cls._construct_empty()
        # UserWarning: Pydantic serializer warnings:
        # New V2 warning, but seems harmless for just printing the schema
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            cls.model_construct(**full_dict).to_yaml(
                output_path, with_comments=True, indent_per_level=indent_per_level
            )

    @classmethod
    def _construct_empty(cls):
        """Construct a model with all fields filled in.

        Uses defaults, or `None` for required fields.

        The .construct is a pydantic method to disable validation
        https://docs.pydantic.dev/usage/models/#creating-models-without-validation
        But Fields without a default value don't get filled in
        so first, we manually make a dict of all the fields with None values
        then we update it with the default-filled values
        """
        all_none_vals = dict(
            zip(cls.model_json_schema()["properties"].keys(), repeat(None))
        )
        all_none_vals.update(cls.model_construct().model_dump())
        return all_none_vals

    def _to_yaml_obj(self, by_alias: bool = True) -> CommentedMap:
        # Make the YAML object to add comments to
        # We can't just do `dumps` for some reason, need a stream
        y = YAML()
        ss = StringIO()
        y.dump(json.loads(self.model_dump_json(by_alias=by_alias)), ss)
        return y.load(ss.getvalue())


def _add_comments(
    loaded_yaml: CommentedMap,
    schema: dict,
    indent: int = 0,
    definitions: Optional[dict] = None,
    # variable specifying how much to indent per level
    indent_per_level: int = 2,
):
    """Add comments above each YAML field using the pydantic model schema."""
    # Definitions are in schemas that contain nested pydantic Models
    defs = schema.get("$defs") if definitions is None else definitions

    for key, val in schema["properties"].items():
        reference = ""
        # Get sub-schema if it exists
        if "$ref" in val:
            # At top level, example is 'outputs': {'$ref': '#/defs/Outputs'}
            reference = val["$ref"]
        elif "allOf" in val:
            # within 'defs', it looks like
            #  'allOf': [{'$ref': '#/defs/HalfWindow'}]
            reference = val["allOf"][0]["$ref"]

        ref_key = reference.split("/")[-1]
        if ref_key:  # The current property is a reference to something else
            if "enum" in defs[ref_key]:  # type: ignore[index]
                # This is just an Enum, not a sub schema.
                # Overwrite the value with the referenced value
                val = defs[ref_key]  # type: ignore[index] # noqa: PLW2901
            else:
                # The reference is a sub schema, so we need to recurse
                sub_schema = defs[ref_key]  # type:ignore[index]
                # Get the sub-model
                sub_loaded_yaml = loaded_yaml[key]
                # recurse on the sub-model
                _add_comments(
                    sub_loaded_yaml,
                    sub_schema,
                    indent=indent + indent_per_level,
                    definitions=defs,
                    indent_per_level=indent_per_level,
                )
                continue

        # add each description along with the type information
        desc = "\n".join(
            textwrap.wrap(
                f"{val['description']}.",
                width=90,
                subsequent_indent=" " * indent_per_level,
            )
        )
        if "anyOf" in val:
            #   'anyOf': [{'type': 'string'}, {'type': 'null'}],
            # Join the options with a pipe, like Python types
            type_str = " | ".join(d["type"] for d in val["anyOf"])
            type_str.replace("null", "None")
        elif "const" in val:
            type_str = val["const"]
        else:
            type_str = val["type"]
        type_line = f"\n  Type: {type_str}."
        choices = f"\n  Options: {val['enum']}." if "enum" in val else ""

        # Combine the description/type/choices as the YAML comment
        comment = f"{desc}{type_line}{choices}"
        comment = comment.replace("..", ".")  # Remove double periods

        # Prepend the required label for fields that are required
        is_required = key in schema.get("required", [])
        if is_required:
            comment = "REQUIRED: " + comment

        # This method comes from here
        # https://yaml.readthedocs.io/en/latest/detail.html#round-trip-including-comments
        loaded_yaml.yaml_set_comment_before_after_key(key, comment, indent=indent)



import argparse
import logging
from enum import Enum
from os import fspath
from pathlib import Path
from typing import Sequence

from osgeo import gdal
from tqdm.contrib.concurrent import thread_map

from dolphin._types import PathOrStr

gdal.UseExceptions()

logger = logging.getLogger(__name__)

DEFAULT_LEVELS = [4, 8, 16, 32, 64]


class Resampling(Enum):
    """GDAL resampling algorithm."""

    NEAREST = "nearest"
    AVERAGE = "average"


class ImageType(Enum):
    """Types of images produced by dolphin."""

    UNWRAPPED = "unwrapped"
    INTERFEROGRAM = "interferogram"
    CORRELATION = "correlation"
    CONNCOMP = "conncomp"
    PS = "ps"


IMAGE_TYPE_TO_RESAMPLING = {
    ImageType.UNWRAPPED: Resampling.AVERAGE,
    ImageType.INTERFEROGRAM: Resampling.AVERAGE,
    ImageType.CORRELATION: Resampling.AVERAGE,
    ImageType.CONNCOMP: Resampling.NEAREST,
    # No max in resampling, yet, which would be best
    # https://github.com/OSGeo/gdal/issues/3683
    ImageType.PS: Resampling.AVERAGE,
}


def create_image_overviews(
    file_path: Path | str,
    levels: Sequence[int] = DEFAULT_LEVELS,
    image_type: ImageType | None = None,
    resampling: Resampling | None = None,
    external: bool = False,
    compression: str = "LZW",
):
    """Add GDAL compressed overviews to an existing file.

    Parameters
    ----------
    file_path : Path
        Path to the file to process.
    levels : Sequence[int]
        List of overview levels to add.
        Default = [4, 8, 16, 32, 64]
    image_type : ImageType, optional
        If provided, looks up the default resampling algorithm
        most appropriate for this type of raster.
    resampling : str or Resampling
        GDAL resampling algorithm for overviews. Required
        if not specifying `image_type`.
    external : bool, default = False
        Use external overviews (.ovr files).
    compression: str, default = "LZW"
        Compression algorithm to use for overviews.
        See https://gdal.org/programs/gdaladdo.html for options.

    """
    if image_type is None and resampling is None:
        raise ValueError("Must provide `image_type` or `resampling`")
    if image_type is not None:
        resampling = IMAGE_TYPE_TO_RESAMPLING[ImageType(image_type)]
    else:
        resampling = Resampling(resampling)

    flags = gdal.GA_Update if not external else gdal.GA_ReadOnly
    ds = gdal.Open(fspath(file_path), flags)
    if ds.GetRasterBand(1).GetOverviewCount() > 0:
        logger.info("%s already has overviews. Skipping.", file_path)
        return

    gdal.SetConfigOption("COMPRESS_OVERVIEW", compression)
    gdal.SetConfigOption("GDAL_NUM_THREADS", "2")
    ds.BuildOverviews(resampling.value, levels)


def create_overviews(
    file_paths: Sequence[PathOrStr],
    levels: Sequence[int] = DEFAULT_LEVELS,
    image_type: ImageType | None = None,
    resampling: Resampling = Resampling.AVERAGE,
    max_workers: int = 5,
) -> None:
    """Process many files to add GDAL overviews and compression.

    Parameters
    ----------
    file_paths : Sequence[PathOrStr]
        Sequence of file paths to process.
    levels : Sequence[int]
        Sequence of overview levels to add.
        Default = [4, 8, 16, 32, 64]
    image_type : ImageType, optional
        If provided, looks up the default resampling algorithm
    resampling : str or Resampling
        GDAL resampling algorithm for overviews. Required
        if not specifying `image_type`.
    max_workers : int, default = 5
        Number of parallel threads to run.

    """
    thread_map(
        lambda file_path: create_image_overviews(
            Path(file_path),
            levels=list(levels),
            image_type=image_type,
            resampling=resampling,
        ),
        file_paths,
        max_workers=max_workers,
    )


def run():
    """Add compressed GDAL overviews to files."""
    parser = argparse.ArgumentParser(
        description="Add compressed GDAL overviews to files.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("file_paths", nargs="+", type=str, help="Path to files")
    parser.add_argument(
        "--levels",
        "-l",
        nargs="*",
        default=[4, 8, 16, 32, 64],
        type=int,
        help="Overview levels to add.",
    )
    parser.add_argument(
        "--resampling",
        "-r",
        default=Resampling("nearest"),
        choices=[r.value for r in Resampling],
        type=Resampling,
        help="Resampling algorithm to use when building overviews",
    )
    parser.add_argument(
        "--max-workers",
        "-n",
        default=5,
        type=int,
        help="Number of parallel files to process",
    )

    args = parser.parse_args()

    # Convert resampling argument from string to Resampling Enum
    resampling_enum = Resampling(args.resampling)

    create_overviews(
        file_paths=args.file_paths,
        levels=args.levels,
        resampling=resampling_enum,
        max_workers=args.max_workers,
    )



"""Main module to provide command line interface to the workflows."""

import sys

from .cli import main

# https://docs.python.org/3/library/__main__.html#packaging-considerations
# allows `python -m dolphin` to work
sys.exit(main())



#!/usr/bin/env python
from __future__ import annotations

import json
import logging
import warnings
from datetime import date, datetime
from os import fspath
from pathlib import Path
from typing import Optional, Sequence

import numpy as np
from opera_utils import get_dates
from pydantic import BaseModel, Field, field_validator, model_validator

from dolphin._types import DateOrDatetime, Filename
from dolphin.io import DEFAULT_DATETIME_FORMAT
from dolphin.utils import format_dates

logger = logging.getLogger(__name__)


class BaseStack(BaseModel):
    """Base class for mini- and full stack classes."""

    file_list: list[Filename] = Field(
        ...,
        description="List of SLC filenames in the ministack.",
    )
    dates: list[Sequence[datetime]] = Field(
        ...,
        description=(
            "List of date sequences, one for each SLC in the ministack. "
            "Each item is a list/tuple of datetime.date or datetime.datetime objects, "
            "as returned by [opera_utils.get_dates][]."
        ),
    )
    is_compressed: list[bool] = Field(
        ...,
        description=(
            "List of booleans indicating whether each SLC is compressed or real."
        ),
    )
    file_date_fmt: str = Field(
        DEFAULT_DATETIME_FORMAT,
        description="Format string for the dates/datetimes in the ministack filenames.",
    )
    output_folder: Path = Field(
        Path(),
        description="Folder/location where ministack will write outputs to.",
    )
    reference_idx: int = Field(
        0,
        description="Index of the SLC to use as reference during phase linking",
    )

    model_config = {
        # For the `Filename, so it can handle the `GeneralPath` protocol`
        # https://github.com/pydantic/pydantic/discussions/5767
        "arbitrary_types_allowed": True
    }

    @field_validator("dates", mode="before")
    @classmethod
    def _check_if_not_tuples(cls, v):
        if isinstance(v[0], (date, datetime)):
            v = [[d] for d in v]
        return v

    @model_validator(mode="after")
    def _check_lengths(self):
        if len(self.file_list) == 0:
            msg = "Cannot create empty ministack"
            raise ValueError(msg)
        elif len(self.file_list) == 1:
            warnings.warn("Creating ministack with only one SLC", stacklevel=2)
        if len(self.file_list) != len(self.is_compressed):
            lengths = f"{len(self.file_list)} and {len(self.is_compressed)}"
            msg = f"file_list and is_compressed must be the same length: Got {lengths}"
            raise ValueError(msg)
        if len(self.dates) != len(self.file_list):
            lengths = f"{len(self.dates)} and {len(self.file_list)}"
            msg = f"dates and file_list must be the same length. Got {lengths}"
            raise ValueError(msg)
        return self

    # @model_validator(mode="after")
    # def _check_unset_reference_date(self):
    #     if self.reference_date == _DUMMY_DATE:
    #         ref_date = self.dates[self.reference_idx][0]
    #         logger.debug("No reference date provided, using first date: %s", ref_date)
    #         self.reference_date = ref_date
    #     return self
    @property
    def reference_date(self):
        """Date of the reference phase of the stack."""
        # Note this works for either a length-1 tuple (real SLC), or for
        # the compressed SLC formate (ref, start, end)
        return self.dates[self.reference_idx][0]

    @property
    def full_date_range(self) -> tuple[DateOrDatetime, DateOrDatetime]:
        """Full date range of all SLCs in the ministack."""
        return (self.reference_date, self.dates[-1][-1])

    @property
    def full_date_range_str(self) -> str:
        """Full date range of the ministack as a string, e.g. '20210101_20210202'.

        Includes both compressed + normal SLCs in the range.
        """
        return format_dates(*self.full_date_range, fmt=self.file_date_fmt)

    @property
    def first_real_slc_idx(self) -> int:
        """Index of the first real SLC in the ministack."""
        try:
            return np.where(~np.array(self.is_compressed))[0][0]
        except IndexError as e:
            msg = "No real SLCs in ministack"
            raise ValueError(msg) from e

    @property
    def real_slc_date_range(self) -> tuple[DateOrDatetime, DateOrDatetime]:
        """Date range of the real SLCs in the ministack."""
        return (self.dates[self.first_real_slc_idx][0], self.dates[-1][-1])

    @property
    def real_slc_date_range_str(self) -> str:
        """Date range of the real SLCs in the ministack."""
        return format_dates(*self.real_slc_date_range, fmt=self.file_date_fmt)

    @property
    def compressed_slc_file_list(self) -> list[Filename]:
        """List of compressed SLCs for this ministack."""
        return [f for f, is_comp in zip(self.file_list, self.is_compressed) if is_comp]

    @property
    def real_slc_file_list(self) -> list[Filename]:
        """List of real SLCs for this ministack."""
        return [
            f for f, is_comp in zip(self.file_list, self.is_compressed) if not is_comp
        ]

    def get_date_str_list(self) -> list[str]:
        """Get a formatted string for each date/date tuple in the ministack."""
        # Should either be like YYYYMMDD, or YYYYMMDD_YYYYMMDD_YYYYMMDD
        return [format_dates(*d, fmt=self.file_date_fmt) for d in self.dates]

    def __rich_repr__(self):
        yield "file_list", self.file_list
        yield "dates", self.dates
        yield "is_compressed", self.is_compressed
        yield "reference_date", self.reference_date
        yield "file_date_fmt", self.file_date_fmt
        yield "output_folder", self.output_folder
        yield "reference_idx", self.reference_idx


class CompressedSlcInfo(BaseModel):
    """Class for holding attributes about one compressed SLC."""

    reference_date: datetime = Field(
        ...,
        description=(
            "Reference date for understanding output interferograms. Note that this may"
            " be different from `start_date` (the first real SLC which  was used in the"
            " compression)."
        ),
    )
    start_date: datetime = Field(
        ..., description="Datetime of the first real SLC used in the compression."
    )
    end_date: datetime = Field(
        ..., description="Datetime of the last real SLC used in the compression."
    )

    real_slc_file_list: list[Filename] | None = Field(
        None,
        description="List of real SLC filenames in the ministack.",
    )
    real_slc_dates: list[datetime] | None = Field(
        None,
        description=(
            "List of date sequences, one for each SLC in the ministack. "
            "Each item is a list/tuple of datetime.date or datetime.datetime objects."
        ),
    )
    compressed_slc_file_list: list[Filename] | None = Field(
        None,
        description="List of compressed SLC filenames in the ministack.",
    )
    file_date_fmt: str = Field(
        DEFAULT_DATETIME_FORMAT,
        description="Format string for the dates/datetimes in the ministack filenames.",
    )
    filename_template: str = Field(
        "compressed_{date_str}.tif",
        description="Template for creating filenames from the CCSLC date triplet.",
    )
    output_folder: Path = Field(
        Path(),
        description="Folder/location where ministack will write outputs to.",
    )

    model_config = {
        # For the `Filename, so it can handle the `GeneralPath` protocol`
        # https://github.com/pydantic/pydantic/discussions/5767
        "arbitrary_types_allowed": True
    }

    @field_validator("real_slc_dates", mode="before")
    @classmethod
    def _untuple_dates(cls, v):
        """Make the dates not be tuples/lists of datetimes."""
        if v is None:
            return v
        out = []
        for item in v:
            if hasattr(item, "__iter__"):
                # Make sure they didn't pass more than 1 date, implying
                # a compressed SLC
                # assert len(item) == 1
                if isinstance(item, str):
                    out.append(item)
                elif len(item) > 1:
                    msg = f"Cannot pass multiple dates for a compressed SLC. Got {item}"
                    raise ValueError(msg)
                else:
                    out.append(item[0])
            else:
                out.append(item)
        return out

    @model_validator(mode="after")
    def _check_lengths(self):
        if self.real_slc_dates is None or self.real_slc_file_list is None:
            return self
        rlen = len(self.real_slc_file_list)
        clen = len(self.real_slc_dates)
        if rlen != clen:
            lengths = f"{rlen} and {clen}"
            msg = (
                "real_slc_file_list and real_slc_dates must be the same length. "
                f"Got {lengths}"
            )
            raise ValueError(msg)
        return self

    @property
    def dates(self) -> tuple[DateOrDatetime, DateOrDatetime, DateOrDatetime]:
        """Alias for the (reference, start, end) date triplet."""
        return (self.reference_date, self.start_date, self.end_date)

    @property
    def real_date_range(self) -> tuple[DateOrDatetime, DateOrDatetime]:
        """Date range of the real SLCs in the ministack."""
        return self.start_date, self.end_date

    @property
    def filename(self) -> str:
        """Create filename using a template with '{date_str}` in the name."""
        date_str = format_dates(*self.dates, fmt=self.file_date_fmt)
        return self.filename_template.format(date_str=date_str)

    @property
    def path(self) -> Path:
        """The path of the compressed SLC for this ministack."""
        return self.output_folder / self.filename

    def write_metadata(
        self, domain: str = "DOLPHIN", output_file: Optional[Filename] = None
    ):
        """Write the metadata to the compressed SLC file.

        Parameters
        ----------
        domain : str, optional
            Domain to write the metadata to, by default "DOLPHIN".
        output_file : Optional[Filename], optional
            Path to the file to write the metadata to, by default None.
            If None, will use `self.path`.

        """
        from dolphin.io import set_raster_metadata

        out = self.path if output_file is None else Path(output_file)
        if not out.exists():
            msg = f"Must create {out} before writing metadata"
            raise FileNotFoundError(msg)

        set_raster_metadata(
            out,
            metadata=self.model_dump(mode="json"),
            domain=domain,
        )

    @classmethod
    def from_filename(
        cls, filename: Filename, date_fmt: str = DEFAULT_DATETIME_FORMAT
    ) -> CompressedSlcInfo:
        """Parse just the dates from a compressed SLC filename."""
        try:
            ref, start, end = get_dates(filename, fmt=date_fmt)[0:3]
        except IndexError as e:
            msg = f"{filename} does not have 3 dates like {date_fmt}"
            raise ValueError(msg) from e
        output_folder = Path(filename).parent
        fname = Path(filename).name

        # Use whichever filename was passed in by replacing the dates we parsed
        _date_str = format_dates(ref, start, end, fmt=date_fmt)
        filename_template = fname.replace(_date_str, "{date_str}")
        return cls(
            reference_date=ref,
            start_date=start,
            end_date=end,
            output_folder=output_folder,
            file_date_fmt=date_fmt,
            filename_template=filename_template,
        )

    @classmethod
    def from_file_metadata(cls, filename: Filename) -> CompressedSlcInfo:
        """Try to parse the CCSLC metadata from `filename`."""
        from dolphin.io import get_raster_metadata

        if not Path(filename).exists():
            raise FileNotFoundError(filename)

        domains = ["DOLPHIN", ""]
        for domain in domains:
            gdal_md = get_raster_metadata(filename, domain=domain)
            if not gdal_md:
                continue
            else:
                break
        else:
            msg = f"Could not find metadata in {filename}"
            raise ValueError(msg)
        # GDAL can write it weirdly and mess up the JSON
        cleaned = {}
        for k, v in gdal_md.items():
            try:
                # Swap the single quotes for double quotes to parse lists
                cleaned[k] = json.loads(v.replace("'", '"'))
            except json.JSONDecodeError:
                cleaned[k] = v
        # Parse the date/file lists from the metadata
        out = cls.model_validate(cleaned)
        # Overwrite the `output_folder` part- we may have moved it since
        # writing the metadata
        out.output_folder = Path(filename).parent
        return out

    def __fspath__(self):
        return fspath(self.path)


class MiniStackInfo(BaseStack):
    """Class for holding attributes about one mini-stack of SLCs.

    Used for planning the processing of a batch of SLCs.
    """

    def get_compressed_slc_info(self) -> CompressedSlcInfo:
        """Get the compressed SLC which will come from this ministack.

        Excludes the existing compressed SLCs during the compression.
        """
        real_slc_files: list[Filename] = []
        real_slc_dates: list[Sequence[DateOrDatetime]] = []
        comp_slc_files: list[Filename] = []
        for f, d, is_comp in zip(self.file_list, self.dates, self.is_compressed):
            if is_comp:
                comp_slc_files.append(f)
            else:
                real_slc_files.append(f)
                real_slc_dates.append(d)

        return CompressedSlcInfo(
            reference_date=self.reference_date,
            start_date=real_slc_dates[0][0],
            end_date=real_slc_dates[-1][0],
            real_slc_file_list=real_slc_files,
            real_slc_dates=real_slc_dates,
            compressed_slc_file_list=comp_slc_files,
            file_date_fmt=self.file_date_fmt,
            output_folder=self.output_folder,
        )


class MiniStackPlanner(BaseStack):
    """Class for planning the processing of batches of SLCs."""

    max_num_compressed: int = 5

    def plan(self, ministack_size: int) -> list[MiniStackInfo]:
        """Create a list of ministacks to be processed."""
        if ministack_size < 2:
            msg = "Cannot create ministacks with size < 2"
            raise ValueError(msg)

        output_ministacks: list[MiniStackInfo] = []

        # Start of with any compressed SLCs that are passed in
        compressed_slc_infos: list[CompressedSlcInfo] = []
        for f in self.compressed_slc_file_list:
            # TODO: will we ever actually need to read the old metadata here?
            # # Note: these must actually exist to be used!
            # compressed_slc_infos.append(CompressedSlcInfo.from_file_metadata(f))
            compressed_slc_infos.append(CompressedSlcInfo.from_filename(f))

        # Solve each ministack using current chunk (and the previous compressed SLCs)
        ministack_starts = range(
            self.first_real_slc_idx, len(self.file_list), ministack_size
        )

        for full_stack_idx in ministack_starts:
            cur_slice = slice(full_stack_idx, full_stack_idx + ministack_size)
            cur_files = list(self.file_list[cur_slice]).copy()
            cur_dates = list(self.dates[cur_slice]).copy()

            # Read compressed*.tif files and if they do not exist use the compressed*.h5
            comp_slc_files = [c.path for c in compressed_slc_infos]
            # Add the existing compressed SLC files to the start, but
            # limit the num comp slcs `max_num_compressed`
            cur_comp_slc_files = comp_slc_files[-self.max_num_compressed :]
            combined_files = cur_comp_slc_files + cur_files

            combined_dates = [
                c.dates for c in compressed_slc_infos[-self.max_num_compressed :]
            ] + cur_dates

            num_ccslc = len(cur_comp_slc_files)
            combined_is_compressed = num_ccslc * [True] + list(
                self.is_compressed[cur_slice]
            )
            # If there are any compressed SLCs, set the reference to the last one
            try:
                last_compressed_idx = np.where(combined_is_compressed)[0]
                reference_idx = last_compressed_idx[-1]
            except IndexError:
                reference_idx = 0

            # Make the current ministack output folder using the start/end dates
            new_date_str = format_dates(
                cur_dates[0][0], cur_dates[-1][-1], fmt=self.file_date_fmt
            )
            cur_output_folder = self.output_folder / new_date_str
            cur_ministack = MiniStackInfo(
                file_list=combined_files,
                dates=combined_dates,
                is_compressed=combined_is_compressed,
                reference_idx=reference_idx,
                output_folder=cur_output_folder,
                reference_date=self.reference_date,
                # TODO: we'll need to alter logic here if we dont fix
                # reference_idx=0, since this will change the reference date
            )

            output_ministacks.append(cur_ministack)
            cur_comp_slc = cur_ministack.get_compressed_slc_info()
            compressed_slc_infos.append(cur_comp_slc)

        return output_ministacks



from concurrent.futures import FIRST_EXCEPTION, Future, ThreadPoolExecutor, wait
from typing import Protocol, Sequence

from numpy.typing import ArrayLike
from tqdm.auto import tqdm

from dolphin.utils import DummyProcessPoolExecutor

from ._blocks import iter_blocks
from ._readers import StackReader
from ._writers import DatasetWriter

__all__ = ["BlockProcessor", "process_blocks"]


class BlockProcessor(Protocol):
    """Protocol for a block-wise processing function.

    Reads a block of data from each reader, processes it, and returns the result
    as an array-like object.
    """

    def __call__(
        self, readers: Sequence[StackReader], rows: slice, cols: slice
    ) -> tuple[ArrayLike, slice, slice]: ...


def process_blocks(
    readers: Sequence[StackReader],
    writer: DatasetWriter,
    func: BlockProcessor,
    block_shape: tuple[int, int] = (512, 512),
    num_threads: int = 5,
):
    """Perform block-wise processing over blocks in `readers`, writing to `writer`.

    Used to read and process a stack of rasters in parallel, setting up a queue
    of results for the `writer` to save.

    Note that the parallelism happens using a `ThreadPoolExecutor`, so `func` should
    be a function which releases the GIL during computation (e.g. using numpy).
    """
    shape = readers[0].shape[-2:]
    slices = list(iter_blocks(shape, block_shape=block_shape))

    pbar = tqdm(total=len(slices))

    # Define the callback to write the result to an output DatasetWrite
    def write_callback(fut: Future):
        data, rows, cols = fut.result()
        writer[..., rows, cols] = data
        pbar.update()

    Executor = ThreadPoolExecutor if num_threads > 1 else DummyProcessPoolExecutor
    futures: set[Future] = set()
    with Executor(num_threads) as exc:
        for rows, cols in slices:
            future = exc.submit(func, readers=readers, rows=rows, cols=cols)
            future.add_done_callback(write_callback)
            futures.add(future)

        while futures:
            done, futures = wait(futures, timeout=1, return_when=FIRST_EXCEPTION)
            for future in done:
                e = future.exception()
                if e is not None:
                    raise e



import datetime
import shutil
import subprocess
from pathlib import Path

import numpy as np
from opera_utils import get_dates
from osgeo import gdal

from dolphin import io, shp
from dolphin._types import HalfWindow, Strides
from dolphin.phase_link import _core, covariance, simulate
from dolphin.stack import MiniStackPlanner
from dolphin.workflows import sequential

# Shared for all tests
HALF_WINDOW = HalfWindow(11, 5)
STRIDES = Strides(3, 6)
HALF_WINDOW_DICT = {"x": HALF_WINDOW.x, "y": HALF_WINDOW.y}
STRIDES_DICT = {"x": STRIDES.x, "y": STRIDES.y}

SHAPE = (512, 512)

# monkeypatch the old names for single covariance computation
if hasattr(covariance, "estimate_stack_covariance_cpu"):
    # The old one wants a dict for half window/strides

    def f(slc_stack, half_window, strides, neighbor_arrays):
        half_window_dict = {"x": half_window.x, "y": half_window.y}
        strides_dict = {"x": strides.x, "y": strides.y}

        return covariance.estimate_stack_covariance_cpu(
            slc_stack,
            half_window=half_window_dict,
            strides=strides_dict,
            neighbor_arrays=neighbor_arrays,
        )

    covariance.estimate_stack_covariance = f  # type: ignore[assignment]


def _make_slc_samples(shape=SHAPE):
    """Create some sample SLC data."""
    # Make it as large as the biggest test
    cov_mat, _ = simulate.simulate_coh(
        num_acq=30,
        Tau0=72,
        gamma_inf=0.3,
        gamma0=0.99,
        add_signal=True,
        signal_std=0.01,
    )
    return simulate.simulate_neighborhood_stack(
        cov_mat, neighbor_samples=np.prod(shape)
    )


class CovarianceBenchmark:
    """Benchmark results for covariance matrix formation."""

    # https://asv.readthedocs.io/en/v0.6.1/writing_benchmarks.html#parameterized-benchmarks
    # Parameterize by the number of SLCs
    params = [10, 20, 30]
    param_names = ["nslc"]

    def setup_cache(self):
        # Run the several-second generation of samples once in setup_cache
        # https://asv.readthedocs.io/en/v0.6.1/writing_benchmarks.html
        np.save("slc_samples.npy", _make_slc_samples())

    def setup(self, nslc):
        self.slc_samples = np.load("slc_samples.npy")[:nslc, :]
        self.slc_stack = self.slc_samples.reshape((nslc, *SHAPE))

    def time_covariance_single(self, nslc):
        # Test one covariance matrix on part of the samples
        covariance.coh_mat_single(self.slc_samples[:, :200])

    def time_covariance_stack(self, nslc):
        covariance.estimate_stack_covariance(
            self.slc_stack, half_window=HALF_WINDOW, strides=STRIDES
        )

    def peakmem_covariance_stack(self, nslc):
        covariance.estimate_stack_covariance(
            self.slc_stack, half_window=HALF_WINDOW, strides=STRIDES
        )


class PhaseLinkingBenchmark:
    """Benchmark phase linking algorithms."""

    # (nslc, use_evd)
    params = ([10, 20, 30], [True, False])
    param_names = ["nslc", "use_evd"]

    def setup_cache(self):
        np.save("slc_samples.npy", _make_slc_samples())

    def setup(self, nslc: int, use_evd: bool):
        self.slc_samples = np.load("slc_samples.npy")[:nslc, :]
        self.slc_stack = self.slc_samples.reshape((nslc, *SHAPE))

    def time_phase_link(self, nslc: int, use_evd: bool):
        _core.run_phase_linking(
            self.slc_stack,
            half_window=HALF_WINDOW,
            strides=STRIDES,
            use_evd=use_evd,
        )

    def peakmem_phase_link(self, nslc: int, use_evd: bool):
        _core.run_phase_linking(
            self.slc_stack,
            half_window=HALF_WINDOW,
            strides=STRIDES,
            use_evd=use_evd,
        )


class ShpBenchmark:
    """Benchmark suite for SHP estimation functions."""

    def setup(self):
        self.nslc = 30
        slc_samples = _make_slc_samples()
        slc_stack = slc_samples.reshape((self.nslc, *SHAPE))

        self.amp_mean = np.mean(np.abs(slc_stack), axis=0)
        self.amp_variance = np.var(np.abs(slc_stack), axis=0)

    def time_estimate_neighbors(self):
        shp.estimate_neighbors(
            halfwin_rowcol=(HALF_WINDOW["y"], HALF_WINDOW["x"]),
            alpha=0.001,
            strides=STRIDES,
            mean=self.amp_mean,
            var=self.amp_variance,
            nslc=30,
            method=shp.ShpMethod.GLRT,
        )


def _make_slc_stack(out_path: Path, shape: tuple[int, int, int] = (10, 1024, 1024)):
    sigma = 0.5
    slc_stack = np.random.normal(0, sigma, size=shape) + 1j * np.random.normal(
        0, sigma, size=shape
    )
    slc_stack = slc_stack.astype(np.complex64)

    start_date = datetime.datetime(2022, 1, 1)
    slc_date_list = []
    dt = datetime.timedelta(days=1)
    for i in range(len(slc_stack)):
        slc_date_list.append(start_date + i * dt)

    shape = slc_stack.shape
    # Write to a file
    driver = gdal.GetDriverByName("GTiff")
    out_path.mkdir(exist_ok=True)
    name_template = out_path / "{date}.slc.tif"

    file_list = []
    for cur_date, cur_slc in zip(slc_date_list, slc_stack):
        fname = str(name_template).format(date=cur_date.strftime("%Y%m%d"))
        file_list.append(Path(fname))
        ds = driver.Create(fname, shape[-1], shape[-2], 1, gdal.GDT_CFloat32)
        ds.GetRasterBand(1).WriteArray(cur_slc)
        ds = None

    return file_list, slc_date_list


class SingleMinistackBenchmark:
    """Benchmark block-by-block phase linking on a stack of rasters."""

    def setup_cache(self):
        _make_slc_stack(Path("slcs"))

    def setup(self):
        # TODO: If we change the dates/size, we'll need a way to remove the last run
        output_folder = Path("pl")
        if output_folder.exists():
            shutil.rmtree(output_folder, ignore_errors=True)

        output_folder.mkdir()
        v_file = output_folder / "stack.vrt"

        subprocess.run("vmtouch -e .", shell=True)  # noqa: PLW1510
        self.slc_file_list = sorted(Path("slcs").glob("20*.slc.tif"))
        assert (
            len(self.slc_file_list) > 0
        ), f"No SLC files found: {list(Path('slcs').glob('*'))}"
        self.dates = [get_dates(f) for f in self.slc_file_list]

        io.VRTStack(self.slc_file_list, outfile=Path("pl") / "stack.vrt")
        self.v_file = v_file

    def time_single_ministack(self):
        ministack_planner = MiniStackPlanner(
            file_list=self.slc_file_list,
            dates=self.dates,
            is_compressed=[False] * len(self.slc_file_list),
            output_folder=Path("pl"),
        )
        # We're using "sequential", but just making it one ministack
        ministack_size = len(self.slc_file_list)
        sequential.run_wrapped_phase_sequential(
            slc_vrt_file=self.v_file,
            ministack_planner=ministack_planner,
            ministack_size=ministack_size,
            half_window=HALF_WINDOW,
            strides=STRIDES,
            block_shape=(512, 512),
            # use_evd=cfg.phase_linking.use_evd,
            # n_workers=cfg.worker_settings.n_workers,
        )



"""Module for handling blocked/decimated input and output."""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Iterator, Optional

from dolphin._types import HalfWindow, Strides
from dolphin.utils import compute_out_shape

# 1. iterate without overlap over output, decimated array
#   - Start with an offset, shrink the size, both by `half_window//strides`
#   - since we want every one wot be full, just skipping the incomplete blocks
# 2. map the current output pixels to their input locations (using strides)
# 3. pad the input block (using half window)

__all__ = [
    "BlockIndices",
    "StridedBlockManager",
    "iter_blocks",
]


logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class BlockIndices:
    """Class holding slices for 2D array access."""

    row_start: int
    row_stop: Optional[int]  # Can be None if we want slice(0, None)
    col_start: int
    col_stop: Optional[int]

    @classmethod
    def from_slices(cls, row_slice: slice, col_slice: slice) -> BlockIndices:
        return cls(
            row_start=row_slice.start,
            row_stop=row_slice.stop,
            col_start=col_slice.start,
            col_stop=col_slice.stop,
        )

    @property
    def row_slice(self) -> slice:
        return slice(self.row_start, self.row_stop)

    @property
    def col_slice(self) -> slice:
        return slice(self.col_start, self.col_stop)

    # Allows user to unpack like (row_slice, col_slice) = block_index
    def __iter__(self):
        return iter((self.row_slice, self.col_slice))


def iter_blocks(
    arr_shape: tuple[int, int],
    block_shape: tuple[int, int],
    overlaps: tuple[int, int] = (0, 0),
    start_offsets: tuple[int, int] = (0, 0),
    end_margin: tuple[int, int] = (0, 0),
) -> Iterator[BlockIndices]:
    """Create a generator to get indexes for accessing blocks of a raster.

    Parameters
    ----------
    arr_shape : tuple[int, int]
        (num_rows, num_cols), full size of array to access
    block_shape : tuple[int, int]
        (height, width), size of blocks to load
    overlaps : tuple[int, int], default = (0, 0)
        (row_overlap, col_overlap), number of pixels to re-include from
        the previous block after sliding
    start_offsets : tuple[int, int], default = (0, 0)
        Offsets from top left to start reading from
    end_margin : tuple[int, int], default = (0, 0)
        Margin to avoid at the bottom/right of array

    Yields
    ------
    BlockIndices
        Iterator of BlockIndices, which can be unpacked into
        (slice(row_start, row_stop), slice(col_start, col_stop))

    Examples
    --------
        >>> list(iter_blocks((180, 250), (100, 100)))
        [BlockIndices(row_start=0, row_stop=100, col_start=0, col_stop=100), \
BlockIndices(row_start=0, row_stop=100, col_start=100, col_stop=200), \
BlockIndices(row_start=0, row_stop=100, col_start=200, col_stop=250), \
BlockIndices(row_start=100, row_stop=180, col_start=0, col_stop=100), \
BlockIndices(row_start=100, row_stop=180, col_start=100, col_stop=200), \
BlockIndices(row_start=100, row_stop=180, col_start=200, col_stop=250)]
        >>> list(map(tuple, iter_blocks((180, 250), (100, 100), overlaps=(10, 10))))
        [(slice(0, 100, None), slice(0, 100, None)), (slice(0, 100, None), \
slice(90, 190, None)), (slice(0, 100, None), slice(180, 250, None)), \
(slice(90, 180, None), slice(0, 100, None)), (slice(90, 180, None), \
slice(90, 190, None)), (slice(90, 180, None), slice(180, 250, None))]

    """
    total_rows, total_cols = arr_shape
    height, width = block_shape
    row_overlap, col_overlap = overlaps
    start_row_offset, start_col_offset = start_offsets
    last_row = total_rows - end_margin[0]
    last_col = total_cols - end_margin[1]

    if height is None:
        height = total_rows
    if width is None:
        width = total_cols

    # Check we're not moving backwards with the overlap:
    if row_overlap >= height and height != total_rows:
        msg = f"{row_overlap = } must be less than block height {height}"
        raise ValueError(msg)
    if col_overlap >= width and width != total_cols:
        msg = f"{col_overlap = } must be less than block width {width}"
        raise ValueError(msg)

    # Set up the iterating indices
    cur_row = start_row_offset
    cur_col = start_col_offset
    while cur_row < last_row:
        while cur_col < last_col:
            row_stop = min(cur_row + height, last_row)
            col_stop = min(cur_col + width, last_col)
            # yield (slice(cur_row, row_stop), slice(cur_col, col_stop))
            yield BlockIndices(cur_row, row_stop, cur_col, col_stop)

            cur_col += width
            if cur_col < last_col:  # dont bring back if already at edge
                cur_col -= col_overlap

        cur_row += height
        if cur_row < last_row:
            cur_row -= row_overlap
        cur_col = start_col_offset  # reset back to the starting offset


def get_full_res_trim(
    in_no_pad_block: BlockIndices, in_block: BlockIndices
) -> BlockIndices:
    in_no_pad_rows, in_no_pad_cols = in_no_pad_block
    in_rows, in_cols = in_block
    trim_full_col = slice(
        in_no_pad_cols.start - in_cols.start, in_no_pad_cols.stop - in_cols.stop
    )
    trim_full_row = slice(
        in_no_pad_rows.start - in_rows.start, in_no_pad_rows.stop - in_rows.stop
    )
    return BlockIndices.from_slices(trim_full_row, trim_full_col)


def get_slice_length(s: slice, data_size: int = 1_000_000):
    """Get the size of a slice of data.

    Uses `slice.indices` to avoid making any dummy data.
    Assumes that 1. data is larger than slice, and 2. size is
    less than `data_size`.
    """
    return len(range(*s.indices(data_size)))


def pad_block(in_block: BlockIndices, margins: tuple[int, int]) -> BlockIndices:
    """Pad `in_block` by the (row_margin, col_margin) pixels in `margins`.

    Will clip the row/column slice starts to be >= 0.

    Parameters
    ----------
    in_block : BlockIndices
        Slices original array block.
    margins : dict[int, int]
        Number of pixels to extend `in_block` in the (row, col) directions.
        The margins are subtracted from the beginning and added to the end,
        so the block size grows by (2 * row_margin, 2 * col_margin)

    Returns
    -------
    BlockIndices
        Output slices for larger block

    Raises
    ------
    ValueError
        If the block is too small to be padded by the given margins
        (leads to start < 0)

    """
    r_margin, c_margin = margins
    r_slice, c_slice = in_block
    if r_slice.start - r_margin < 0:
        msg = f"{r_slice = }, but {r_margin = }"
        raise ValueError(msg)
    if c_slice.start - c_margin < 0:
        msg = f"{c_slice = }, but {c_margin = }"
        raise ValueError(msg)
    return BlockIndices(
        # max(r_slice.start - r_margin, 0),
        r_slice.start - r_margin,
        r_slice.stop + r_margin,
        # max(c_slice.start - c_margin, 0),
        c_slice.start - c_margin,
        c_slice.stop + c_margin,
    )


def dilate_block(in_block: BlockIndices, strides: Strides) -> BlockIndices:
    """Grow slices in `BlockIndices` to fit a larger array.

    This is to undo the "stride"/decimation index changes, so
     we can go from smaller, strided array indices to the original.

    Assumes in in_block is the smaller one which has been made
    by taking `strides` from the larger block.

    Parameters
    ----------
    in_block : BlockIndices
        Slices for an smaller, strided array.
    strides : Strides or tuple[int, int]
        Decimation factor in y and c which was used for `in_block`

    Returns
    -------
    BlockIndices
        Output slices for larger array

    """
    row_strides, col_strides = strides
    row_start = in_block.row_start * row_strides
    col_start = in_block.col_start * col_strides
    row_stop = None if in_block.row_stop is None else in_block.row_stop * row_strides
    col_stop = None if in_block.col_stop is None else in_block.col_stop * col_strides
    return BlockIndices(row_start, row_stop, col_start, col_stop)


@dataclass
class StridedBlockManager:
    """Class to handle slicing/trimming overlapping blocks with strides."""

    arr_shape: tuple[int, int]
    """(row, col) of the full-res 2D image"""
    block_shape: tuple[int, int]
    """(row, col) size of each input block to operate on at one time"""
    strides: Strides = field(default_factory=lambda: Strides(1, 1))
    """Decimation/downsampling factor in y/row and x/column direction"""
    half_window: HalfWindow = field(default_factory=lambda: HalfWindow(0, 0))
    """For multi-looking iterations, size of the (full-res) half window
    in y/row and x/column direction.
    Used to find `overlaps` between blocks and `start_offset`/`end_margin` for
    `iter_blocks`."""

    def iter_blocks(
        self,
    ) -> Iterator[
        tuple[BlockIndices, BlockIndices, BlockIndices, BlockIndices, BlockIndices]
    ]:
        """Iterate over the input/output block indices.

        Yields
        ------
        output_block : BlockIndices
            The current slices for the output raster
        out_trim : BlockIndices
            Slices to use on a processed output block to remove nodata border pixels.
            These may be relative (e.g. slice(1, -1)), not absolute like `output_block`.
        input_block : BlockIndices
            Slices used to load the full-res input data
        input_no_padding : BlockIndices
            Slices which point to the position within the full-res data without padding
        input_trim : BlockIndices
            Slices to use on full-res input block to remove padding from half-window.

        """
        out_trim = self.get_trimming_block()
        for out_block in self.iter_outputs():
            input_no_padding = dilate_block(out_block, strides=self.strides)
            input_block = pad_block(input_no_padding, margins=self.input_padding_shape)
            input_trim = get_full_res_trim(input_no_padding, input_block)
            yield (out_block, out_trim, input_block, input_no_padding, input_trim)

    def _get_out_nodata_size(self, direction: str) -> int:
        half_win = getattr(self.half_window, direction)
        stride = getattr(self.strides, direction)
        return round(half_win / stride)

    @property
    def output_shape(self) -> tuple[int, int]:
        return compute_out_shape(self.arr_shape, self.strides)

    @property
    def out_block_shape(self) -> tuple[int, int]:
        return compute_out_shape(self.block_shape, self.strides)

    @property
    def input_padding_shape(self) -> tuple[int, int]:
        """Amount of extra padding the input blocks need.

        Depends on the window size, and the strides.
        """
        return (
            self.strides.y * self._get_out_nodata_size("y"),
            self.strides.x * self._get_out_nodata_size("x"),
        )

    @property
    def output_margin(self) -> tuple[int, int]:
        """The output margins that we ignore while iterating.

        Depends on the half window and strides.

        The `half_window` is in full-res (input) coordinates (which would be the
        amount to skip with no striding), so the output margin size is smaller
        """
        return (self._get_out_nodata_size("y"), self._get_out_nodata_size("x"))

    def iter_outputs(self) -> Iterator[BlockIndices]:
        yield from iter_blocks(
            arr_shape=self.output_shape,
            block_shape=self.out_block_shape,
            overlaps=(0, 0),  # We're not overlapping in the *output* grid
            start_offsets=self.output_margin,
            end_margin=self.output_margin,
        )

    def get_trimming_block(self) -> BlockIndices:
        """Compute the slices which trim output nodata values.

        When the BlockIndex gets dilated (using `strides`) and padded (using
        `half_window`), the result will have nodata around the edges.
        The size of the nodata pixels in the full-res block is just
            (half_window['y'], half_window['x'])
        In the output (strided) coordinates, the number of nodata pixels is
        shrunk by how many strides are taken.

        Note that this is independent of which block we're on; the number of
        nodata pixels on the border is always the same.
        """
        row_nodata = self._get_out_nodata_size("y")
        col_nodata = self._get_out_nodata_size("x")
        # Extra check if we have no trimming to do: use slice(0, None)
        row_end = -row_nodata if row_nodata > 0 else None
        col_end = -col_nodata if col_nodata > 0 else None
        return BlockIndices(row_nodata, row_end, col_nodata, col_end)



import zipfile
from pathlib import Path

import numpy as np
import pytest

from dolphin import Bbox, io, masking


@pytest.fixture()
def mask_files(tmp_path):
    """Make series of files offset in lat/lon."""
    shape = (9, 9)

    all_masked = np.zeros(shape, dtype="uint8")
    file_list = []
    for i in range(shape[0]):
        fname = tmp_path / f"mask_{i}.tif"
        arr = all_masked.copy()
        # make first third, then middle, then end "good" pixels
        rows = slice(3 * i, 3 * (i + 1))
        arr[rows, :] = 1
        io.write_arr(arr=arr, output_name=fname)
        file_list.append(Path(fname))

    return file_list


@pytest.mark.parametrize("convention", [None, masking.MaskConvention.ZERO_IS_NODATA])
def test_combine_mask_files(mask_files, convention):
    output_file = mask_files[0].parent / "combined.tif"
    masking.combine_mask_files(
        mask_files=mask_files,
        output_file=output_file,
        input_conventions=convention,
    )
    expected = np.zeros((9, 9), dtype="uint8")
    np.testing.assert_array_equal(expected, io.load_gdal(output_file))


def test_load_mask_as_numpy(mask_files):
    arr = masking.load_mask_as_numpy(mask_files[0])
    expected = np.ones((9, 9), dtype=bool)
    expected[:3] = False
    np.testing.assert_array_equal(arr, expected)


@pytest.fixture
def like_filename_zipped():
    return Path(__file__).parent / "data/dummy_like.tif.zip"


def test_bounds(tmp_path, like_filename_zipped):
    # Unzip to tmp_path
    with zipfile.ZipFile(like_filename_zipped, "r") as zip_ref:
        zip_ref.extractall(tmp_path)

    # Get the path of the extracted TIF file
    extracted_tif = tmp_path / "dummy_like.tif"

    output_filename = tmp_path / "mask_bounds.tif"
    bounds = Bbox(
        left=-122.90334860812246,
        bottom=51.7323987260125,
        right=-122.68416491724179,
        top=51.95333755674119,
    )
    masking.create_bounds_mask(
        bounds, like_filename=extracted_tif, output_filename=output_filename
    )
    # Check result
    mask = io.load_gdal(output_filename)
    assert (mask[1405:3856, 9681:12685] == 1).all()
    # WGS84 box is not a box in UTM
    assert (mask[:1400, :] == 0).all()
    assert (mask[4000:, :] == 0).all()
    assert (mask[:, :9500] == 0).all()
    assert (mask[:, 13000:] == 0).all()



from __future__ import annotations

import logging
from typing import Optional

import numpy as np
from numpy.typing import ArrayLike

from dolphin.workflows import ShpMethod

from . import _glrt, _ks

logger = logging.getLogger(__name__)

__all__ = ["estimate_neighbors"]


def estimate_neighbors(
    *,
    halfwin_rowcol: tuple[int, int],
    alpha: float,
    strides: Optional[dict[str, int]] = None,
    mean: Optional[ArrayLike] = None,
    var: Optional[ArrayLike] = None,
    nslc: Optional[int] = None,
    amp_stack: Optional[ArrayLike] = None,
    is_sorted: bool = False,
    method: ShpMethod = ShpMethod.GLRT,
    prune_disconnected: bool = False,
) -> np.ndarray:
    """Estimate the statistically similar neighbors of each pixel.

    GLRT method on the [@Parizzi2011AdaptiveInSARStack].
    Assumes Rayleigh distributed amplitudes ([@Siddiqui1962ProblemsConnectedRayleigh]).

    Parameters
    ----------
    halfwin_rowcol : Tuple[int, int]
        Half window dimensions as a tuple (rows, columns).
    alpha : float
        Significance level (0 < alpha < 1).
    strides : dict[str, int], optional
        Strides for the x and y dimensions, by default {"x": 1, "y": 1}.
    mean : Optional[ArrayLike], optional
        Mean of the amplitude stack, by default None.
    var : Optional[ArrayLike], optional
        Variance of the amplitude stack, by default None.
    nslc : Optional[int], optional
        Number of samples, by default None.
    amp_stack : Optional[ArrayLike], optional
        Amplitude stack, by default None.
    is_sorted : bool, optional
        Whether the amplitude stack is sorted (if passed), by default False.
    method : ShpMethod, optional
        Method used for estimation, by default ShpMethod.GLRT.
    prune_disconnected : bool, default=False
        If True, keeps only SHPs that are 8-connected to the current pixel.
        Otherwise, any pixel within the window may be considered an SHP, even
        if it is not directly connected.

    Returns
    -------
    Optional[np.ndarray]
        Array of estimated statistically similar neighbors.

    Raises
    ------
    ValueError
        - nslc is not provided for GLRT method
        - amp_stack is not provided for the KS method.
        - `method` not a valid `ShpMethod`

    """
    import numba

    if strides is None:
        strides = {"x": 1, "y": 1}
    logger.debug(f"NUMBA THREADS: {numba.get_num_threads()}")

    if method == ShpMethod.RECT:
        # No estimation needed
        neighbor_arrays = None
    elif method.lower() == ShpMethod.GLRT:
        logger.debug("Estimating SHP neighbors using GLRT")
        if nslc is None:
            msg = "`nslc` must be provided for GLRT method"
            raise ValueError(msg)
        if mean is None:
            mean = np.mean(amp_stack, axis=0)
        if var is None:
            var = np.var(amp_stack, axis=0)
        neighbor_arrays = _glrt.estimate_neighbors(
            mean=mean,
            var=var,
            halfwin_rowcol=halfwin_rowcol,
            strides=strides,
            nslc=nslc,
            alpha=alpha,
            prune_disconnected=prune_disconnected,
        )
    elif method.lower() == ShpMethod.KS:
        if amp_stack is None:
            msg = "amp_stack must be provided for KS method"
            raise ValueError(msg)
        logger.debug("Estimating SHP neighbors using KS test")
        neighbor_arrays = _ks.estimate_neighbors(
            amp_stack=amp_stack,
            halfwin_rowcol=halfwin_rowcol,
            strides=strides,
            alpha=alpha,
            is_sorted=is_sorted,
        )
    else:
        msg = f"SHP method {method} is not implemented"
        raise ValueError(msg)

    return neighbor_arrays



import pytest
import rasterio as rio

from dolphin._overviews import (
    ImageType,
    Resampling,
    create_image_overviews,
    create_overviews,
)

# Dataset has no geotransform, gcps, or rpcs. The identity matrix will be returned.
pytestmark = pytest.mark.filterwarnings(
    "ignore::rasterio.errors.NotGeoreferencedWarning",
    "ignore:.*io.FileIO.*:pytest.PytestUnraisableExceptionWarning",
)


def get_overviews(filename, band=1):
    with rio.open(filename) as src:
        return src.overviews(band)


def test_create_image_overviews(list_of_gtiff_ifgs):
    f = list_of_gtiff_ifgs[0]
    assert len(get_overviews(f)) == 0
    create_image_overviews(f, image_type=ImageType.INTERFEROGRAM)
    assert len(get_overviews(f)) > 0


def test_create_image_overviews_envi(list_of_envi_ifgs):
    f = list_of_envi_ifgs[0]
    assert len(get_overviews(f)) == 0
    create_image_overviews(f, image_type=ImageType.INTERFEROGRAM)
    assert len(get_overviews(f)) > 0


@pytest.mark.parametrize("resampling", list(Resampling))
def test_resamplings(list_of_gtiff_ifgs, resampling):
    f = list_of_gtiff_ifgs[0]
    assert len(get_overviews(f)) == 0
    create_image_overviews(f, resampling=resampling)
    assert len(get_overviews(f)) > 0


@pytest.mark.parametrize("levels", [[2, 4], [4, 8, 18]])
def test_levels(list_of_gtiff_ifgs, levels):
    f = list_of_gtiff_ifgs[0]
    assert len(get_overviews(f)) == 0
    create_image_overviews(f, resampling="nearest", levels=levels)
    assert len(get_overviews(f)) == len(levels)


def test_create_overviews(list_of_gtiff_ifgs):
    create_overviews(list_of_gtiff_ifgs, image_type=ImageType.INTERFEROGRAM)
    assert all(len(get_overviews(f)) > 0 for f in list_of_gtiff_ifgs)



"""Module for computing phase similarity between complex interferogram pixels.

Uses metric from @[Wang2022AccuratePersistentScatterer] for similarity.
"""

import logging
from pathlib import Path
from typing import Callable, Literal, Sequence

import numba
import numpy as np
from numpy.typing import ArrayLike

from dolphin._types import PathOrStr

logger = logging.getLogger(__name__)


@numba.njit(nogil=True)
def phase_similarity(x1: ArrayLike, x2: ArrayLike):
    """Compute the similarity between two complex 1D vectors."""
    n = len(x1)
    out = 0.0
    for i in range(n):
        out += np.real(x1[i] * np.conj(x2[i]))
    return out / n


def median_similarity(
    ifg_stack: ArrayLike, search_radius: int, mask: ArrayLike | None = None
):
    """Compute the median similarity of each pixel and its neighbors.

    Resulting similarity matches Equation (5) of @[Wang2022AccuratePersistentScatterer]

    Parameters
    ----------
    ifg_stack : ArrayLike
        3D stack of complex interferograms.
        Shape is (n_ifg, rows, cols)
    search_radius: int
        maximum radius (in pixels) to search for neighbors when comparing each pixel.
    mask: ArrayLike (optional)
        Array of mask from True/False indicating whether to ignore the pixel

    Returns
    -------
    np.ndarray
        2D array (shape (rows, cols)) of the median similarity at each pixel.

    """
    return _create_loop_and_run(
        ifg_stack=ifg_stack,
        search_radius=search_radius,
        mask=mask,
        func=np.nanmedian,
    )


def max_similarity(
    ifg_stack: ArrayLike, search_radius: int, mask: ArrayLike | None = None
):
    """Compute the maximum similarity of each pixel and its neighbors.

    Resulting similarity matches Equation (6) of @[Wang2022AccuratePersistentScatterer]

    Parameters
    ----------
    ifg_stack : ArrayLike
        3D stack of complex interferograms.
        Shape is (n_ifg, rows, cols)
    search_radius: int
        maximum radius (in pixels) to search for neighbors when comparing each pixel.
    mask: ArrayLike (optional)
        Array of mask from True/False indicating whether to ignore the pixel

    Returns
    -------
    np.ndarray
        2D array (shape (rows, cols)) of the maximum similarity for any neighrbor
        at a pixel.

    """
    return _create_loop_and_run(
        ifg_stack=ifg_stack,
        search_radius=search_radius,
        mask=mask,
        func=np.nanmax,
    )


def _create_loop_and_run(
    ifg_stack: ArrayLike,
    search_radius: int,
    mask: ArrayLike,
    func: Callable[[ArrayLike], np.ndarray],
):
    n_ifg, rows, cols = ifg_stack.shape
    if not np.iscomplexobj(ifg_stack):
        raise ValueError("ifg_stack must be complex")

    unit_ifgs = np.exp(1j * np.angle(ifg_stack))
    out_similarity = np.zeros((rows, cols), dtype="float32")
    if mask is None:
        mask = np.ones((rows, cols), dtype="bool")

    if mask.shape != (rows, cols):
        raise ValueError(f"{ifg_stack.shape = }, but {mask.shape = }")

    idxs = get_circle_idxs(search_radius)
    loop_func = _make_loop_function(func)
    return loop_func(unit_ifgs, idxs, mask, out_similarity)


def _make_loop_function(
    summary_func: Callable[[ArrayLike], np.ndarray],
):
    """Create a JIT-ed function for some summary of the neighbors's similarity.

    E.g.: for median similarity, call

        median_sim = _make_loop_function(np.median)
    """

    @numba.njit(nogil=True, parallel=True)
    def _masked_median_sim_loop(
        ifg_stack: np.ndarray,
        idxs: np.ndarray,
        mask: np.ndarray,
        out_similarity: np.ndarray,
    ) -> np.ndarray:
        """Loop over each pixel, make a masked phase similarity to its neighbors."""
        _, rows, cols = ifg_stack.shape

        num_compare_pixels = len(idxs)
        # Buffer to hold all comparison during the parallel loop
        cur_sim = np.zeros((rows, cols, num_compare_pixels))

        for r0 in numba.prange(rows):
            for c0 in range(cols):
                # Get the current pixel
                m0 = mask[r0, c0]
                if not m0:
                    continue
                x0 = ifg_stack[:, r0, c0]
                cur_sim_vec = cur_sim[r0, c0]
                count = 0

                # compare to all pixels in the circle around it
                for i_idx in range(num_compare_pixels):
                    ir, ic = idxs[i_idx]
                    # Clip to the image bounds
                    r = max(min(r0 + ir, rows - 1), 0)
                    c = max(min(c0 + ic, cols - 1), 0)
                    if r == r0 and c == c0:
                        continue

                    # Check for a pixel to ignore
                    if not mask[r, c]:
                        continue

                    x = ifg_stack[:, r, c]
                    # cur_sim_vec[count] = w * phase_similarity(x0, x)
                    cur_sim_vec[count] = phase_similarity(x0, x)
                    count += 1
                    # Assuming `summary_func` is nan-aware
                out_similarity[r0, c0] = summary_func(cur_sim_vec[:count])
        return out_similarity

    return _masked_median_sim_loop


def get_circle_idxs(
    max_radius: int, min_radius: int = 0, sort_output: bool = True
) -> np.ndarray:
    """Get the relative indices of neighboring pixels in a circle.

    Adapted from c++ version of `psps` package:
    https://github.com/UT-Radar-Interferometry-Group/psps/blob/a15d458817fe7d06a6edaa0b3208ea78bc4782e7/src/cpp/similarity.cpp#L16
    """
    # using the mid-point circle drawing algorithm to search for neighboring PS pixels
    # # code adapted from "https://www.geeksforgeeks.org/mid-point-circle-drawing-algorithm/"
    visited = np.zeros((max_radius, max_radius), dtype=bool)
    visited[0][0] = True

    indices = []
    for r in range(1, max_radius):
        x = r
        y = 0
        p = 1 - r
        if r > min_radius:
            indices.append([r, 0])
            indices.append([-r, 0])
            indices.append([0, r])
            indices.append([0, -r])

        visited[r][0] = True
        visited[0][r] = True
        # flag > 0 means there are holes between concentric circles
        flag = 0
        while x > y:
            # do not need to fill holes
            if flag == 0:
                y += 1
                if p <= 0:
                    # Mid-point is inside or on the perimeter
                    p += 2 * y + 1
                else:
                    # Mid-point is outside the perimeter
                    x -= 1
                    p += 2 * y - 2 * x + 1

            else:
                flag -= 1

            # All the perimeter points have already been visited
            if x < y:
                break

            while not visited[x - 1][y]:
                x -= 1
                flag += 1

            visited[x][y] = True
            visited[y][x] = True
            if r > min_radius:
                indices.append([x, y])
                indices.append([-x, -y])
                indices.append([x, -y])
                indices.append([-x, y])

                if x != y:
                    indices.append([y, x])
                    indices.append([-y, -x])
                    indices.append([y, -x])
                    indices.append([-y, x])

            if flag > 0:
                x += 1

    if sort_output:
        # Sorting makes it run faster, better data access patterns
        return np.array(sorted(indices))
    else:
        # Indices run from middle outward
        return np.array(indices)


def create_similarities(
    ifg_file_list: Sequence[PathOrStr],
    output_file: PathOrStr,
    search_radius: int = 7,
    sim_type: Literal["median", "max"] = "median",
    block_shape: tuple[int, int] = (512, 512),
    num_threads: int = 5,
    add_overviews: bool = True,
):
    """Create a similarity raster from as stack of ifg files.

    Parameters
    ----------
    ifg_file_list : Sequence[PathOrStr]
        Paths to input interferograms
    output_file : PathOrStr
        Output raster path
    search_radius : int, optional
        Maximum radius to search for pixels, by default 7
    sim_type : str, optional
        Type of similarity function to run, by default "median"
        Choices: "median", "max"
    block_shape : tuple[int, int], optional
        Size of blocks to process at one time from `ifg_file_list`
        by default (512, 512)
    num_threads : int, optional
        Number of parallel blocks to process, by default 5
    add_overviews : bool, optional
        Whether to create overviews in `output_file` by default True

    """
    from dolphin._overviews import Resampling, create_image_overviews
    from dolphin.io import BackgroundRasterWriter, VRTStack, process_blocks

    if sim_type == "median":
        sim_function = median_similarity
    elif sim_type == "max":
        sim_function = max_similarity
    else:
        raise ValueError(f"Unrecognized {sim_type = }")

    zero_block = np.zeros(block_shape, dtype="float32")

    def calc_sim(readers, rows, cols):
        block = readers[0][:, rows, cols]
        if np.sum(block) == 0 or np.isnan(block).all():
            return zero_block[rows, cols], rows, cols

        out_avg = sim_function(ifg_stack=block, search_radius=search_radius)
        logger.debug(f"{rows = }, {cols = }, {block.shape = }, {out_avg.shape = }")
        return out_avg, rows, cols

    out_dir = Path(output_file).parent
    reader = VRTStack(ifg_file_list, outfile=out_dir / "sim_inputs.vrt")
    if reader.dtype not in (np.complex64, np.complex128):
        raise ValueError(
            f"ifg_file_list must be complex interferograms. Got {reader.dtype}"
        )

    writer = BackgroundRasterWriter(
        output_file,
        like_filename=ifg_file_list[0],
        dtype="float32",
        driver="GTiff",
    )
    process_blocks(
        [reader],
        writer,
        func=calc_sim,
        block_shape=block_shape,
        num_threads=num_threads,
    )
    writer.notify_finished()

    if add_overviews:
        logger.info("Creating overviews for unwrapped images")
        create_image_overviews(Path(output_file), resampling=Resampling.AVERAGE)



from __future__ import annotations

import glob
import logging
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Optional

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
    field_validator,
    model_validator,
)

from dolphin import __version__ as _dolphin_version
from dolphin._types import Bbox
from dolphin.io import DEFAULT_HDF5_OPTIONS, DEFAULT_TIFF_OPTIONS

from ._enums import ShpMethod
from ._yaml_model import YamlModel

logger = logging.getLogger(__name__)

__all__ = [
    "HalfWindow",
    "InputOptions",
    "OutputOptions",
    "WorkerSettings",
    "PsOptions",
    "PhaseLinkingOptions",
    "InterferogramNetwork",
    "TimeseriesOptions",
]


class PsOptions(BaseModel, extra="forbid"):
    """Options for the PS pixel selection portion of the workflow."""

    _directory: Path = PrivateAttr(Path("PS"))
    _output_file: Path = PrivateAttr(Path("PS/ps_pixels.tif"))
    _amp_dispersion_file: Path = PrivateAttr(Path("PS/amp_dispersion.tif"))
    _amp_mean_file: Path = PrivateAttr(Path("PS/amp_mean.tif"))

    amp_dispersion_threshold: float = Field(
        0.25,
        description="Amplitude dispersion threshold to consider a pixel a PS.",
        gt=0.0,
    )


class HalfWindow(BaseModel, extra="forbid"):
    """Class to hold half-window size for multi-looking during phase linking."""

    x: int = Field(11, description="Half window size (in pixels) for x direction", gt=0)
    y: int = Field(5, description="Half window size (in pixels) for y direction", gt=0)

    def to_looks(self):
        """Convert (x, y) half-window size to (row, column) looks."""
        return 2 * self.y + 1, 2 * self.x + 1

    @classmethod
    def from_looks(cls, row_looks: int, col_looks: int):
        """Create a half-window from looks."""
        return cls(x=col_looks // 2, y=row_looks // 2)


class PhaseLinkingOptions(BaseModel, extra="forbid"):
    """Configurable options for wrapped phase estimation."""

    _directory: Path = PrivateAttr(Path("linked_phase"))
    ministack_size: int = Field(
        10, description="Size of the ministack for sequential estimator.", gt=1
    )
    max_num_compressed: int = Field(
        5,
        description=(
            "Maximum number of compressed images to use in sequential estimator."
            " If there are more ministacks than this, the earliest CCSLCs will be"
            " left out of the later stacks."
        ),
        gt=0,
    )
    half_window: HalfWindow = HalfWindow()
    use_evd: bool = Field(
        False, description="Use EVD on the coherence instead of using the EMI algorithm"
    )

    beta: float = Field(
        0.00,
        description=(
            "Beta regularization parameter for correlation matrix inversion. 0 is no"
            " regularization."
        ),
        ge=0.0,
        le=1.0,
    )
    shp_method: ShpMethod = ShpMethod.GLRT
    shp_alpha: float = Field(
        0.005,
        description="Significance level (probability of false alarm) for SHP tests.",
        gt=0.0,
        lt=1.0,
    )
    baseline_lag: Optional[int] = Field(
        None,
        gt=0,
        description=(
            "StBAS parameter to include only nearest-N interferograms for"
            "phase linking. A `baseline_lag` of `n` will only include the closest"
            "`n` interferograms. `baseline_line` must be positive."
        ),
    )


class InterferogramNetwork(BaseModel, extra="forbid"):
    """Options to determine the type of network for interferogram formation.

    If no parameters passed, uses single-reference network with `reference_idx=0`.
    """

    _directory: Path = PrivateAttr(Path("interferograms"))

    reference_idx: Optional[int] = Field(
        None,
        description=(
            "For single-reference network: Index of the reference image in the network"
        ),
    )
    max_bandwidth: Optional[int] = Field(
        None,
        description="Max `n` to form the nearest-`n` interferograms by index.",
        ge=1,
    )
    max_temporal_baseline: Optional[int] = Field(
        None,
        description="Maximum temporal baseline of interferograms.",
        ge=0,
    )
    indexes: Optional[list[tuple[int, int]]] = Field(
        None,
        description=(
            "For manual-index network: list of (ref_idx, sec_idx) defining the"
            " interferograms to form."
        ),
    )

    # validation
    @model_validator(mode="after")
    def _check_zero_parameters(self) -> InterferogramNetwork:
        ref_idx = self.reference_idx
        max_bw = self.max_bandwidth
        max_tb = self.max_temporal_baseline
        indexes = self.indexes
        # Check if more than one has been set:
        if ref_idx is None and max_bw is None and max_tb is None and indexes is None:
            logger.debug(
                "No network configuration options were set. Using single-reference."
            )
            self.reference_idx = 0
        return self


class TimeseriesOptions(BaseModel, extra="forbid"):
    """Options for inversion/time series fitting."""

    _directory: Path = PrivateAttr(Path("timeseries"))
    _velocity_file: Path = PrivateAttr(Path("timeseries/velocity.tif"))
    run_inversion: bool = Field(
        True,
        description=(
            "Whether to run the inversion step after unwrapping, if more than "
            " a single-reference network is used."
        ),
    )
    reference_point: Optional[tuple[int, int]] = Field(
        None,
        description=(
            "Reference point (row, col) used if performing a time series inversion. "
            "If not provided, a point will be selected from a consistent connected "
            "component with low amplitude dispersion."
        ),
    )

    run_velocity: bool = Field(
        True,
        description="Run the velocity estimation from the phase time series.",
    )
    correlation_threshold: float = Field(
        0.2,
        description="Pixels with correlation below this value will be masked out.",
        ge=0.0,
        le=1.0,
    )


class WorkerSettings(BaseModel, extra="forbid"):
    """Settings for controlling CPU/GPU settings and parallelism."""

    gpu_enabled: bool = Field(
        False,
        description="Whether to use GPU for processing (if available)",
    )
    threads_per_worker: int = Field(
        1,
        ge=1,
        description=(
            "Number of threads to use per worker. This sets the OMP_NUM_THREADS"
            " environment variable in each python process."
        ),
    )
    n_parallel_bursts: int = Field(
        default=1,
        ge=1,
        description=(
            "If processing separate spatial bursts, number of bursts to run in parallel"
            " for wrapped-phase-estimation."
        ),
    )
    block_shape: tuple[int, int] = Field(
        (512, 512),
        description="Size (rows, columns) of blocks of data to load at a time.",
    )


class InputOptions(BaseModel, extra="forbid"):
    """Options specifying input datasets for workflow."""

    subdataset: Optional[str] = Field(
        None,
        description="If passing HDF5/NetCDF files, subdataset to use from CSLC files. ",
    )
    cslc_date_fmt: str = Field(
        "%Y%m%d",
        description="Format of dates contained in CSLC filenames",
    )


class OutputOptions(BaseModel, extra="forbid"):
    """Options for the output size/format/compressions."""

    output_resolution: Optional[dict[str, int]] = Field(
        # {"x": 20, "y": 20},
        # TODO: how to get a blank "x" and "y" in the schema printed instead of nothing?
        None,
        description="Output (x, y) resolution (in units of input data)",
    )
    strides: dict[str, int] = Field(
        {"x": 1, "y": 1},
        description=(
            "Alternative to specifying output resolution: Specify the (x, y) strides"
            " (decimation factor) to perform while processing input. For example,"
            " strides of [4, 2] would turn an input resolution of [5, 10] into an"
            " output resolution of [20, 20]."
        ),
        validate_default=True,
    )
    bounds: Optional[tuple[float, float, float, float]] = Field(
        None,
        description=(
            "Area of interest: [left, bottom, right, top] coordinates. "
            "e.g. `bbox=[-150.2,65.0,-150.1,65.5]`"
        ),
    )
    bounds_epsg: int = Field(
        4326, description="EPSG code for the `bounds` coordinates, if specified."
    )

    hdf5_creation_options: dict = Field(
        DEFAULT_HDF5_OPTIONS,
        description="Options for `create_dataset` with h5py.",
    )
    gtiff_creation_options: list[str] = Field(
        list(DEFAULT_TIFF_OPTIONS),
        description="GDAL creation options for GeoTIFF files",
    )
    add_overviews: bool = Field(
        True,
        description=(
            "Whether to add overviews to the output GeoTIFF files. This will "
            "increase file size, but can be useful for visualizing the data with "
            "web mapping tools. See https://gdal.org/programs/gdaladdo.html for more."
        ),
    )
    overview_levels: list[int] = Field(
        [4, 8, 16, 32, 64],
        description="List of overview levels to create (if `add_overviews=True`).",
    )

    # validators
    @field_validator("bounds", mode="after")
    @classmethod
    def _convert_bbox(cls, bounds):
        if bounds:
            return Bbox(*bounds)
        return bounds

    @field_validator("strides")
    @classmethod
    def _check_strides_against_res(cls, strides, info):
        """Compute the output resolution from the strides."""
        resolution = info.data.get("output_resolution")
        if strides is not None and resolution is not None:
            msg = "Cannot specify both strides and output_resolution."
            raise ValueError(msg)
        elif strides is None and resolution is None:
            msg = "Must specify either strides or output_resolution."
            raise ValueError(msg)

        # Check that the dict has the correct keys
        if strides is not None:
            if set(strides.keys()) != {"x", "y"}:
                msg = "Strides must be a dict with keys 'x' and 'y'"
                raise ValueError(msg)
            # and that the strides are integers
            if not all(isinstance(v, int) for v in strides.values()):
                msg = "Strides must be integers"
                raise ValueError(msg)
        if resolution is not None:
            if set(resolution.keys()) != {"x", "y"}:
                msg = "Resolution must be a dict with keys 'x' and 'y'"
                raise ValueError(msg)
            # and that the resolution is valid, > 0. Can be int or float
            if any(v <= 0 for v in resolution.values()):
                msg = "Resolutions must be > 0"
                raise ValueError(msg)
            # TODO: compute strides from resolution
            msg = "output_resolution not yet implemented. Use `strides`."
            raise NotImplementedError(msg)
        return strides


class WorkflowBase(YamlModel):
    """Base of multiple workflow configuration models."""

    # Paths to input/output files
    input_options: InputOptions = Field(default_factory=InputOptions)

    mask_file: Optional[Path] = Field(
        None,
        description=(
            "Mask file used to ignore low correlation/bad data (e.g water mask)."
            " Convention is 0 for no data/invalid, and 1 for good data. Dtype must be"
            " uint8."
        ),
    )
    work_directory: Path = Field(
        Path(),
        description="Name of sub-directory to use for writing output files",
        validate_default=True,
    )
    keep_paths_relative: bool = Field(
        False,
        description=(
            "Don't resolve filepaths that are given as relative to be absolute."
        ),
    )

    # General workflow metadata
    worker_settings: WorkerSettings = Field(default_factory=WorkerSettings)
    log_file: Optional[Path] = Field(
        default=None,
        description=(
            "Path to output log file (in addition to logging to `stderr`)."
            " Default logs to `dolphin.log` within `work_directory`"
        ),
    )
    creation_time_utc: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
        description="Time the config file was created",
    )

    model_config = ConfigDict(extra="allow")
    _dolphin_version: str = PrivateAttr(_dolphin_version)
    # internal helpers
    # Stores the list of directories to be created by the workflow
    _directory_list: list[Path] = PrivateAttr(default_factory=list)

    def model_post_init(self, __context: Any) -> None:
        """After validation, set up properties for use during workflow run."""
        super().model_post_init(__context)
        # Ensure outputs from workflow steps are within work directory.
        if not self.keep_paths_relative:
            # Save all directories as absolute paths
            self.work_directory = self.work_directory.resolve(strict=False)

    def create_dir_tree(self) -> None:
        """Create the directory tree for the workflow."""
        for d in self._directory_list:
            logger.debug(f"Creating directory: {d}")
            d.mkdir(parents=True, exist_ok=True)


def _read_file_list_or_glob(cls, value):  # noqa: ARG001:
    """Check if the input file list is a glob pattern or a text file.

    If it's a text file, read the lines and return a list of Path objects.
    If it's a string representing a glob pattern, return a list of Path objects.

    Parameters
    ----------
    cls
        pydantic model class
    value : str | Path | list[str] | list[Path]
        Value passed to pydantic model: Input file list.

    """
    if value is None:
        return []

    # Check if they've passed a glob pattern
    if (
        isinstance(value, (list, tuple))
        and (len(value) == 1)
        and glob.has_magic(str(value[0]))
    ):
        value = glob.glob(str(value[0]))
    elif isinstance(value, (str, Path)):
        v_path = Path(value)

        # Check if it's a newline-delimited list of input files
        if glob.has_magic(str(value)):
            value = glob.glob(str(value))
        elif v_path.exists() and v_path.is_file():
            filenames = [Path(f) for f in v_path.read_text().splitlines()]

            # If given as relative paths, make them relative to the text file
            parent = v_path.parent
            return [parent / f if not f.is_absolute() else f for f in filenames]
        else:
            msg = f"Input file list {v_path} does not exist or is not a file."
            raise ValueError(msg)

    return list(value)



"""Module for estimating covariance matrices for stacks or single pixels.

Contains for CPU and GPU versions (which will not be available if no GPU).
"""

from __future__ import annotations

from functools import partial
from typing import Optional

import jax.numpy as jnp
import numpy as np
from jax import Array, jit, lax, vmap
from jax.typing import ArrayLike

from dolphin._types import HalfWindow, Strides
from dolphin.utils import compute_out_shape

DEFAULT_STRIDES = Strides(1, 1)

__all__ = ["estimate_stack_covariance", "coh_mat_single"]


# @partial(jit, static_argnames=["half_window", "strides"])
def estimate_stack_covariance(
    slc_stack: ArrayLike,
    half_window: HalfWindow,
    strides: Strides = DEFAULT_STRIDES,
    neighbor_arrays: Optional[np.ndarray] = None,
) -> Array:
    """Estimate the linked phase at all pixels of `slc_stack`.

    Parameters
    ----------
    slc_stack : ArrayLike
        The SLC stack, with shape (n_slc, n_rows, n_cols).
    half_window : tuple[int, int]
        A (named) tuple of (y, x) sizes for the half window.
        The full window size is 2 * half_window + 1 for x, y.
    strides : tuple[int, int], optional
        The (y, x) strides (in pixels) to use for the sliding window.
        By default (1, 1)
    neighbor_arrays : np.ndarray, optional
        The neighbor arrays to use for SHP, shape = (n_rows, n_cols, *window_shape).
        If None, a rectangular window is used. By default None.

    Returns
    -------
    C_arrays : np.ndarray
        The covariance matrix at each pixel, with shape
        (n_rows, n_cols, n_slc, n_slc).

    Raises
    ------
    ValueError
        If `slc_stack` is not complex data.

    """
    if not np.iscomplexobj(slc_stack):
        msg = "The SLC stack must be complex."
        raise ValueError(msg)
    if neighbor_arrays is None:
        rows, cols = slc_stack.shape[1:]
        full_window = (2 * half_window.y + 1, 2 * half_window.x + 1)
        neighbor_arrays = jnp.ones((rows, cols, *full_window), dtype=bool)
    return estimate_stack_covariance_t(
        jnp.moveaxis(slc_stack, 0, -1), half_window, strides, neighbor_arrays
    )


@partial(jit, static_argnames=["half_window", "strides"])
def estimate_stack_covariance_t(
    slc_stack_reshaped: ArrayLike,
    half_window: HalfWindow,
    strides: Strides = DEFAULT_STRIDES,
    neighbor_arrays: Optional[np.ndarray] = None,
) -> Array:
    """Estimate the linked phase at all pixels of `slc_stack_reshaped`."""
    rows, cols, nslc = slc_stack_reshaped.shape

    row_strides = strides.y
    col_strides = strides.x
    half_row = half_window.y
    half_col = half_window.x

    out_rows, out_cols = compute_out_shape((rows, cols), strides)

    in_r_start = row_strides // 2
    in_c_start = col_strides // 2

    if neighbor_arrays is None:
        neighbor_arrays = jnp.ones(
            (out_rows, out_cols, 2 * half_window[0] + 1, 2 * half_window[1] + 1),
            dtype=bool,
        )

    def _process_row_col(out_r, out_c):
        """Get slices for, and process, one pixel's window."""
        in_r = in_r_start + out_r * row_strides
        in_c = in_c_start + out_c * col_strides
        # Get a 3D slice, size (row_window, col_window, nslc)
        slc_window = _get_stack_window_t(
            slc_stack_reshaped, in_r, in_c, half_row, half_col
        )
        # Reshape to be (num_samples, nslc)
        slc_samples = slc_window.reshape(-1, nslc)
        cur_neighbors = neighbor_arrays[out_r, out_c, :, :]
        neighbor_mask = cur_neighbors.ravel()

        return coh_mat_single_t(slc_samples, neighbor_mask=neighbor_mask)

    # Now make a 2D grid of indices to access all output pixels
    out_r_indices, out_c_indices = jnp.meshgrid(
        jnp.arange(out_rows), jnp.arange(out_cols), indexing="ij"
    )

    # Create the vectorized function in 2d
    _process_2d = vmap(_process_row_col)
    # Then in 3d
    _process_3d = vmap(_process_2d)
    return _process_3d(out_r_indices, out_c_indices)


def _get_stack_window_t(
    stack: ArrayLike, r: int, c: int, half_row: int, half_col: int
) -> Array:
    """Dynamically slice the stack at (r, c) with size (2*half_row+1, 2*half_col+1).

    Expected shape of `stack` is (rows, cols, nslc).
    """
    # https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html#out-of-bounds-indexing
    # Note: out of bounds indexing for JAX clamps to the nearest value
    # This is fine as we want to trim the borders anyway, so we don't need the
    # extra checks in utils._get_slices

    # Center the slice on (r, c), so we need the starts to move up/left
    # The upper bound on `clamp` isn't used meaningless here
    r0 = lax.clamp(0, r - half_row, r)
    c0 = lax.clamp(0, c - half_col, c)
    start_indices = (r0, c0, 0)

    # Note: we can't clamp the size using a max size,
    # TypeError: Shapes must be 1D sequences of concrete values of integer type,
    # got  Traced<ShapedArray(int32[],  ...
    rsize = 2 * half_row + 1
    csize = 2 * half_col + 1
    dsize = stack.shape[-1]
    slice_sizes = (rsize, csize, dsize)
    return lax.dynamic_slice(stack, start_indices, slice_sizes)


@jit
def coh_mat_single(
    slc_samples: ArrayLike, neighbor_mask: Optional[ArrayLike] = None
) -> Array:
    """Given (n_slc, n_samps) SLC samples, get the (nslc, nslc) coherence matrix.

    Note this requires `slc_samples` to be transposed from `coh_mat_single`.
    """
    return coh_mat_single_t(jnp.moveaxis(slc_samples, 0, -1), neighbor_mask)


@jit
def coh_mat_single_t(
    slc_samples: ArrayLike, neighbor_mask: Optional[ArrayLike] = None
) -> Array:
    """Given (n_samps, nslc) SLC samples, get the (nslc, nslc) coherence matrix.

    Note this requires `slc_samples` to be transposed from `coh_mat_single`.
    """
    nsamps, _ = slc_samples.shape

    if neighbor_mask is None:
        neighbor_mask = jnp.ones(nsamps, dtype=jnp.bool_)
    valid_samples_mask = ~jnp.isnan(slc_samples)
    combined_mask = valid_samples_mask & neighbor_mask[:, None]

    # Mask the slc samples
    # note that it's not possible to change the size based on the mask
    # https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html#dynamic-shapes
    masked_slc = jnp.where(combined_mask, slc_samples, 0)

    # Compute cross-correlation
    numer = jnp.dot(masked_slc.T, jnp.conj(masked_slc))

    # Compute amplitudes so we normalize the covariance to a coherence matrix
    # a1 is shape (nslc,)
    amp_vec = jnp.sum(jnp.abs(masked_slc) ** 2, axis=0)
    # Form outer product of amplitudes for each slc
    power_mat = amp_vec[:, None] * amp_vec[None, :]
    amp_mat = jnp.sqrt(power_mat)

    # Compute coherence matrix
    # Make the output 0 where the amplitudes are 0 (instead of nan/divide by 0)
    return jnp.where(amp_mat > 1e-6, numer / amp_mat, 0 + 0j)
